import os.path
import re
from typing import Literal, TYPE_CHECKING, Dict

import catboost
import matplotlib.patheffects as pe
import numpy as np
import pandas as pd
import pytorch_lightning as pl
import shap
import wandb

from matplotlib import pyplot as plt
from pytorch_lightning import loggers, profiler
from scipy.sparse import coo_array

from my_tools.utils import torch_sparse_slice
from .data import SparseDataModuleInterface
from .interface import ExplanationMixin
from .metrics import RecommendingMetrics
from .utils import wandb_plt_figure, filter_warnings, save_shap_force_plot, Timer

if TYPE_CHECKING:
    from scipy.sparse import spmatrix
    from .interface import RecommenderModuleBase
    from .lit import LitRecommenderBase
    from .models.cat import CatboostInterface


class WandbWatcher(pl.callbacks.Callback):
    """Log torch parameter values to wandb."""

    def __init__(
        self,
        log_what: Literal["gradients", "parameters", "all"] = "all",
        log_every_n_steps=100,
        log_graph=True,
    ):
        self.watch_kwargs = dict(
            log=log_what,
            log_freq=log_every_n_steps,
            log_graph=log_graph,
        )
        # Wandb errors if you try to watch same model twice
        self.watch_model_triggered = False

    def setup(self, trainer, pl_module, stage=None):
        if self.watch_model_triggered:
            return
        if not isinstance(trainer.logger, loggers.WandbLogger):
            raise ValueError("Only wandb logger supports watching model.")
        trainer.logger.watch(pl_module, **self.watch_kwargs)
        self.watch_model_triggered = True

    def teardown(self, trainer, pl_module, stage=None):
        trainer.logger.experiment.unwatch(pl_module)


class RecommendingDataOverviewCallback(pl.callbacks.Callback):
    """Logs to wandb some general information about the data."""

    def __init__(self, explicit: "spmatrix" = None):
        self.explicit = explicit

    def setup(self, trainer, pl_module, stage=None):
        if self.explicit is None:
            self.explicit = pl_module.train_explicit
            self.log_data_overview()

    @staticmethod
    def add_quantiles(
        data,
        cmap="hsv",
        quantiles_to_plot=(0.5, 0.75, 0.9, 0.99),
    ):
        for q in quantiles_to_plot:
            q_value = np.quantile(data, q=q)
            plt.axvline(q_value, ls="--", c=plt.get_cmap(name=cmap)(q))
            plt.text(
                x=q_value,
                y=plt.ylim()[1] * 0.95,
                s=f"{100 * q}%",
                ha="left",
                color="white",
                path_effects=[pe.withStroke(linewidth=2, foreground="black")],
            )

    def log_data_description(self):
        series = pd.Series(
            {
                "n_users": self.explicit.shape[0],
                "n_items": self.explicit.shape[1],
                "density": self.explicit.nnz / np.prod(self.explicit.shape),
            }
        )
        ratings_description = pd.Series(self.explicit.data).describe()
        ratings_description.index = "ratings_" + ratings_description.index
        series = pd.concat([series, ratings_description])
        wandb.log({"Data description": wandb.Table(dataframe=series.to_frame().T)})

    def log_ratings_distributions(self):
        user_reviews_counts = self.explicit.sum(axis=1).A.squeeze()
        item_reviews_counts = self.explicit.sum(axis=0).A.squeeze()
        for what, data in zip(
            ["user", "item"], [user_reviews_counts, item_reviews_counts]
        ):
            with wandb_plt_figure(title=f"Distribution of {what} reviews"):
                plt.xlabel(f"Number of reviews per {what}")
                plt.ylabel("Quantity")
                plt.hist(data, histtype="stepfilled", bins=50)
                self.add_quantiles(data)

    def log_data_overview(self):
        self.log_data_description()
        self.log_ratings_distributions()

        with wandb_plt_figure(title="Distribution of ratings"):
            plt.xlabel("Number of reviews per rating value")
            plt.ylabel("Quantity")
            plt.hist(self.explicit.data, histtype="stepfilled", bins=50)


class WandbCheckpointCallback(pl.callbacks.ModelCheckpoint):
    def __init__(self, *args, artifact_name, aliases=None, **kwargs):
        self.artifact_name = artifact_name
        self.aliases = aliases
        super().__init__(*args, **kwargs)

    def on_fit_end(self, trainer, pl_module):
        if not self.best_model_path:
            self.save_checkpoint(trainer)
        trainer.logger.log_text(
            key="checkpoint", columns=["path"], data=[[self.best_model_path]]
        )
        wandb.log_artifact(
            artifact_or_path=self.best_model_path,
            name=self.artifact_name,
            type="checkpoint",
            aliases=self.aliases,
        )


class CatBoostMetrics(pl.callbacks.Callback):
    """Log some metrics specific to catboost."""

    def __init__(
        self,
        *args,
        plot_dependence=False,
        max_dataset_size_for_complex_shap=1000,
        figsize=(8, 5),
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.plot_dependence = plot_dependence
        self.max_dataset_size_for_complex_shap = max_dataset_size_for_complex_shap
        self.figsize = figsize

    def on_train_end(self, trainer, pl_module: "LitRecommenderBase"):
        self.log_feature_names(model=pl_module.model)
        self.common_on_end(pl_module=pl_module, stage="train")

    def on_validation_end(self, trainer, pl_module: "LitRecommenderBase"):
        self.common_on_end(pl_module=pl_module, stage="val")

    def on_test_end(self, trainer, pl_module: "LitRecommenderBase"):
        self.common_on_end(pl_module=pl_module, stage="test")

    @Timer()
    def common_on_end(self, pl_module, stage: 'Literal["train", "val", "test"]'):
        model: "CatboostInterface" = pl_module.model
        if stage == "train":
            user_item_dataframe = model.train_user_item_dataframe
        elif stage == "val":
            user_item_dataframe = model.build_labeled_user_item_dataframe(
                explicit=pl_module.val_explicit()
            )
        elif stage == "test":
            user_item_dataframe = model.build_labeled_user_item_dataframe(
                explicit=pl_module.test_explicit()
            )
        else:
            raise ValueError(f"Unknown stage {stage}")

        wandb.log(
            {
                f"{stage} dataframe size": float(len(user_item_dataframe)),
                f"shap expected value": model.shap_explainer.expected_value,
            }
        )

        self.log_single_user_features_dataframe_with_ratings_and_explicit(
            model=model, dataframe=user_item_dataframe, stage=stage
        )

        pool = model.pool(user_item_dataframe=user_item_dataframe)
        self.log_feature_importance(catboost_model=model.model, pool=pool, stage=stage)

        shap_kwargs = model.shap_kwargs(user_item_dataframe=user_item_dataframe)

        self.log_shap_summary_plot(**shap_kwargs, stage=stage, figsize=self.figsize)

        if self.plot_dependence:
            self.log_shap_dependence_plots(
                **shap_kwargs, stage=stage, figsize=self.figsize
            )

        if len(user_item_dataframe) > self.max_dataset_size_for_complex_shap:
            user_item_dataframe = user_item_dataframe.sample(
                self.max_dataset_size_for_complex_shap
            ).sort_index()
            shap_kwargs = model.shap_kwargs(user_item_dataframe=user_item_dataframe)

        self.log_shap_force_plot(**shap_kwargs, stage=stage)

        dataframe = model.merge_features(user_item_dataframe)
        pool_kwargs = model.extract_pool_kwargs(dataframe=dataframe)
        postprocessed_pool_kwargs = model.postprocess_pool_kwargs(**pool_kwargs)
        shap_explanation = model.shap_explainer(X=postprocessed_pool_kwargs["data"])
        self.log_shap_heatmap_plot(
            shap_explanation=shap_explanation, stage=stage, figsize=self.figsize
        )

    @staticmethod
    def log_feature_names(model: "CatboostInterface"):
        ...
        # model.cached_indexed_features_dataframe doesn't have all the features -> wrong description
        # cat_features = {k.value: list(v) for k, v in model.cat_features.items()}
        # text_features = {k.value: list(v) for k, v in model.text_features.items()}
        # non_numeric_features = sum(cat_features.values(), start=[]) + sum(
        #     text_features.values(), start=[]
        # )
        # numeric_features = {}
        # for kind in model.FeatureKind:
        #     if (features := model.cached_indexed_features_dataframe(kind)) is not None:
        #         numeric_features[kind.value] = list(
        #             features.columns.difference(non_numeric_features)
        #         )
        # features_description = pd.DataFrame(
        #     dict(
        #         cat_features=cat_features,
        #         text_features=text_features,
        #         numeric_features=numeric_features,
        #     )
        # ).reset_index()
        # wandb.log(dict(features=wandb.Table(dataframe=features_description)))

    @staticmethod
    def log_single_user_features_dataframe_with_ratings_and_explicit(
        model: "CatboostInterface", dataframe: pd.DataFrame, stage: str
    ):
        user_id = dataframe.reset_index()["user_id"].sample(1).values[0]
        offline_dataframe = dataframe.query(f"user_id == @user_id")
        user_explicit = coo_array(
            (
                offline_dataframe["rating"].values,
                (
                    np.zeros(len(offline_dataframe)),
                    offline_dataframe.reset_index()["item_id"].values,
                ),
            ),
            shape=[1, model.n_items],
        )
        online_dataframe = model.build_recommend_user_item_dataframe(
            users_explicit=model.to_torch_coo(user_explicit),
            user_ids=model.fictive_user_ids(1),
            n_recommendations=len(offline_dataframe),
        )
        for kind, dataframe in zip(
            ["offline", "online"], [offline_dataframe, online_dataframe]
        ):
            pool = model.pool(user_item_dataframe=dataframe)
            features_dataframe = model.merge_features(dataframe)
            with filter_warnings(
                action="ignore", category=pd.errors.SettingWithCopyWarning
            ):
                features_dataframe["ratings"] = model.model.predict(pool)
            features_dataframe = features_dataframe.sort_values(
                "ratings", ascending=False
            ).reset_index()
            wandb.log(
                {
                    f"{stage} {kind} user dataframe": wandb.Table(
                        dataframe=features_dataframe
                    )
                }
            )

    @staticmethod
    def log_feature_importance(
        catboost_model: "catboost.CatBoost", pool: "catboost.Pool", stage: str
    ):
        dataframe = catboost_model.get_feature_importance(pool, prettified=True)
        wandb_table = wandb.Table(dataframe=dataframe)
        wandb.log({f"Catboost {stage} feature importance": wandb_table})

    @staticmethod
    def log_shap_force_plot(
        base_value: float, shap_values: np.array, features: pd.DataFrame, stage: str
    ):
        shap_plot = shap.force_plot(
            base_value=base_value,
            shap_values=shap_values,
            features=features,
            out_names="Predicted relevance of items for users",
            text_rotation=0,
        )
        textio = save_shap_force_plot(shap_plot=shap_plot)
        wandb.log({f"shap/{stage} force plot": wandb.Html(textio)})

    @staticmethod
    def log_shap_dependence_plots(
        base_value: float,
        shap_values: np.array,
        features: pd.DataFrame,
        stage: str,
        figsize,
    ):
        for feature_name in list(features):
            with wandb_plt_figure(
                title=f"shap/dependence plot/{stage} " + feature_name, figsize=figsize
            ) as figure:
                shap.dependence_plot(
                    feature_name,
                    shap_values=shap_values,
                    features=features,
                    ax=figure.gca(),
                )

    @staticmethod
    def log_shap_summary_plot(
        base_value: float,
        shap_values: np.array,
        features: pd.DataFrame,
        stage: str,
        figsize,
    ):
        with wandb_plt_figure(title=f"shap/{stage} summary plot", figsize=figsize):
            shap.summary_plot(shap_values=shap_values, features=features)

    @staticmethod
    def log_shap_heatmap_plot(shap_explanation: shap.Explanation, stage: str, figsize):
        with wandb_plt_figure(title=f"shap/{stage} heatmap plot", figsize=figsize):
            shap.plots.heatmap(shap_values=shap_explanation)


class RecommendingMetricsCallback(pl.callbacks.Callback):
    def __init__(
        self,
        k: int = 10,
        every_val_epoch=1,
        every_test_epoch=1,
        assess_online_recommendations=False,
    ):
        self.k = k
        self.every_epoch = dict(val=every_val_epoch, test=every_test_epoch)
        self.assess_online_recommendations = assess_online_recommendations
        self.metrics: "Dict[str, RecommendingMetrics]" = dict()
        self.log_dict = wandb.log

    def skip_epoch(self, current_epoch, stage) -> bool:
        every_epoch = self.every_epoch[stage]
        return every_epoch == 0 or current_epoch % every_epoch

    def setup(self, trainer, pl_module, stage=None):
        if isinstance(trainer.datamodule, SparseDataModuleInterface):
            module = trainer.datamodule
        elif isinstance(pl_module, SparseDataModuleInterface):
            module = pl_module
        else:
            raise ValueError(
                f"One of lightning_module, datamodule must implement "
                f"the {SparseDataModuleInterface.__name__} interface."
            )
        self.log_dict = pl_module.log_dict
        for stage in ["val", "test"]:
            explicit = getattr(module, f"{stage}_explicit")()
            for is_online in [True, False]:
                if explicit is not None and (
                    not is_online or self.assess_online_recommendations
                ):
                    kind = self.kind(stage=stage, is_online=is_online)
                    self.metrics[kind] = RecommendingMetrics(explicit)

    @staticmethod
    def kind(stage: str, is_online: bool) -> str:
        assert stage in ["val", "test"], f"Unknown stage {stage}."
        return stage + "_" + ("online" if is_online else "offline")

    @Timer()
    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        self.common_on_batch_end(pl_module=pl_module, batch=batch, stage="val")

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        self.common_on_batch_end(pl_module=pl_module, batch=batch, stage="test")

    def on_validation_epoch_end(self, trainer, pl_module):
        self.common_epoch_end(stage="val")

    def on_test_epoch_end(self, trainer, pl_module):
        self.common_epoch_end(stage="test")

    def common_on_batch_end(self, pl_module, batch, stage):
        if self.skip_epoch(current_epoch=pl_module.current_epoch, stage=stage):
            return
        user_ids = batch["user_ids"]
        model = pl_module.model
        for is_online in [True, False]:
            kind = self.kind(stage=stage, is_online=is_online)
            if (metrics := self.metrics.get(kind)) is not None:
                if is_online:
                    users_explicit = torch_sparse_slice(
                        model.explicit, row_ids=user_ids
                    )
                    recommendations = model.online_recommend(
                        users_explicit=users_explicit, n_recommendations=self.k
                    )
                else:
                    recommendations = model.recommend(
                        user_ids=user_ids, n_recommendations=self.k
                    )
                metrics_dict = metrics.batch_metrics(
                    user_ids=user_ids, recommendations=recommendations
                )
                self.log_metrics(metrics_dict=metrics_dict, prefix=kind)

    def log_metrics(self, metrics_dict: dict, prefix: str):
        metrics_dict = {prefix + "_" + k: v for k, v in metrics_dict.items()}
        for metric_name in metrics_dict:
            wandb.define_metric(metric_name, summary="mean")
        self.log_dict(metrics_dict)

    def common_epoch_end(self, stage):
        for is_online in [True, False]:
            kind = self.kind(stage=stage, is_online=is_online)
            if (metrics := self.metrics.get(kind)) is not None and len(
                metrics.unique_recommended_items
            ):
                metrics = metrics.finalize_coverage()
                self.log_metrics(metrics, prefix=kind)


class RecommendingExplanationCallback(pl.callbacks.Callback):
    def __init__(
        self,
        user_ids=None,
        users_explicit=None,
        n_recommendations=10,
    ):
        self.user_ids = user_ids
        self.users_explicit = users_explicit
        self.n_recommendations = n_recommendations

    def on_test_epoch_end(self, trainer, pl_module):
        self.my_on_epoch_end(model=pl_module.model, stage="test")

    def on_validation_epoch_end(self, trainer, pl_module):
        self.my_on_epoch_end(model=pl_module.model, stage="val")

    def my_on_epoch_end(
        self,
        model: "RecommenderModuleBase",
        stage: Literal["train", "val", "test", "predict"],
    ):
        if not isinstance(model, ExplanationMixin):
            return
        if self.user_ids is not None:
            for user_id in self.user_ids:
                model.explain_recommendations_for_user(
                    user_id=user_id,
                    n_recommendations=self.n_recommendations,
                    log=True,
                    logging_prefix=f"explanation/{stage} ",
                )
        if self.users_explicit is not None:
            try:
                for user_explicit in self.users_explicit.detach().clone():
                    model.explain_recommendations_for_user(
                        user_explicit=user_explicit[None],
                        n_recommendations=self.n_recommendations,
                        log=True,
                        logging_prefix=f"explanation/{stage} ",
                    )
            except NotImplementedError:
                pass


class WandbSaveCodeOnceCallback(pl.callbacks.Callback):
    def __init__(
        self, root=".", name=None, include_regex=".*\\.py", exclude_regex=None
    ):
        self.root = root
        self.name = name
        self.include_regex = include_regex
        self.exclude_regex = exclude_regex
        self.have_saved_code = False

    def include_fn(self, path_to_code):
        return bool(re.fullmatch(pattern=self.include_regex, string=path_to_code))

    def exclude_fn(self, path_to_code):
        if self.exclude_regex is None:
            return False
        return bool(re.fullmatch(pattern=self.exclude_regex, string=path_to_code))

    def setup(self, trainer, pl_module, stage=None):
        if not self.have_saved_code:
            wandb.run.log_code(
                root=self.root,
                name=self.name,
                include_fn=self.include_fn,
                exclude_fn=self.exclude_fn,
            )
            self.have_saved_code = True


class WandbProfiler(profiler.SimpleProfiler):
    def __init__(self, artifact_name=None, dirpath=None, filename=None, extended=True):
        self.artifact_name = artifact_name or "unnamed"
        super().__init__(
            dirpath=dirpath or "local",
            filename=filename or "profiler_logs",
            extended=extended,
        )

    def get_path(self, stage):
        return os.path.join(self.dirpath, f"{stage}-{self.filename}.txt")

    def teardown(self, stage=None):
        super().teardown(stage=stage)
        path = self.get_path(stage=stage)
        if os.path.exists(path) and wandb.run is not None:
            html = wandb.Html(f"<pre>{open(path).read()}</pre>")
            wandb.log({f"profiler/{stage}": html})
            os.remove(path)


class ParameterStatsLoggerCallback(pl.callbacks.Callback):
    def __init__(self, every_train_batch=1):
        self.every_train_batch = every_train_batch

    @staticmethod
    def parameter_stats(parameter):
        stats = dict(
            numel=float(parameter.numel()),
            mean=parameter.mean(),
            isnan_sum=float(parameter.isnan().sum()),
            abs_max=parameter.abs().max(),
            norm=parameter.norm(),
        )
        if (grad := parameter.grad) is not None:
            stats["grad_norm"] = grad.detach().norm()
        return stats

    def on_train_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, **kwargs
    ):
        if batch_idx % self.every_train_batch == 0:
            for parameter_name, parameter in pl_module.named_parameters():
                for stat_name, stat_value in self.parameter_stats(parameter).items():
                    pl_module.log(f"parameter/{stat_name}_{parameter_name}", stat_value)



import itertools

from typing import Literal, TYPE_CHECKING

import numpy as np
import torch

from torch.utils.data import DataLoader, Dataset, IterableDataset

from my_tools.utils import torch_sparse_slice

if TYPE_CHECKING:
    import pytorch_lightning as pl
    from scipy.sparse import csr_matrix


class SparseDataset(Dataset):
    def __init__(self, explicit: "csr_matrix"):
        self.explicit = explicit

    def __len__(self):
        return self.explicit.shape[0]

    @property
    def shape(self):
        return self.explicit.shape

    def __getitem__(self, indices):
        user_ids = indices.get("user_ids", torch.arange(self.shape[0]))
        item_ids = indices.get("item_ids", torch.arange(self.shape[1]))

        explicit = torch_sparse_slice(self.explicit, user_ids, item_ids)
        return dict(
            explicit=self.pack_sparse_tensor(explicit),
            user_ids=user_ids,
            item_ids=item_ids,
        )

    @staticmethod
    def pack_sparse_tensor(torch_sparse):
        torch_sparse = torch_sparse.coalesce()
        sparse_kwargs = dict(
            indices=torch_sparse.indices(),
            values=torch_sparse.values(),
            size=torch_sparse.size(),
        )
        return sparse_kwargs

    @staticmethod
    def unpack_sparse_kwargs_to_torch_sparse_csr(batch):
        for key, value in batch.items():
            if isinstance(value, dict) and set(value) == {
                "indices",
                "values",
                "size",
            }:
                batch[key] = torch.sparse_coo_tensor(**value).to_sparse_csr()
        return batch


class SparseTensorUnpacker:
    def on_after_batch_transfer(self: "pl.LightningModule", batch, dataloader_idx=0):
        """
        Need to manually pack and then unpack sparse torch matrices because
        they are unsupported in Dataloaders as of the moment of writing.
        """
        return SparseDataset.unpack_sparse_kwargs_to_torch_sparse_csr(batch)


class Sampler:
    def __init__(self, size, batch_size, shuffle=True):
        indices = torch.randperm(size) if shuffle else torch.arange(size)
        self.batch_indices = torch.split(indices, batch_size)

    def __len__(self):
        return len(self.batch_indices)

    def __iter__(self):
        yield from self.batch_indices


class UserSampler(Sampler):
    def __iter__(self):
        yield from ({"user_ids": i} for i in super().__iter__())


class ItemSampler(Sampler):
    def __iter__(self):
        yield from ({"item_ids": i} for i in super().__iter__())


class GridSampler:
    """
    Splits user ids and item ids into chunks, and uses
    cartesian product of these chunked ids to generate batches.
    """

    def __init__(self, dataset_shape, approximate_batch_size, shuffle=True):
        self.dataset_shape = dataset_shape
        self.chunks_per_dim = max(
            1,
            (
                (torch.tensor(dataset_shape).prod() / approximate_batch_size)
                ** (1 / len(dataset_shape))
            )
            .round()
            .int(),
        )
        self.shuffle = shuffle

    def __len__(self):
        return self.chunks_per_dim ** len(self.dataset_shape)

    def __iter__(self):
        batch_indices_per_dimension = [
            torch.randperm(dimension_size).chunk(self.chunks_per_dim)
            for dimension_size in self.dataset_shape
        ]
        numpy_batches = [[j.numpy() for j in i] for i in batch_indices_per_dimension]
        batch_indices_product = itertools.product(*numpy_batches)
        if self.shuffle:
            batch_indices_product = np.array(list(batch_indices_product), dtype=object)
            batch_indices_product = np.random.permutation(batch_indices_product)
        yield from ({"user_ids": i[0], "item_ids": i[1]} for i in batch_indices_product)


class GridIterableDataset(GridSampler, IterableDataset):
    pass


def build_recommending_sampler(
    batch_size,
    n_users,
    n_items,
    sampler_type: Literal["grid", "user", "item"],
    shuffle,
):
    if sampler_type == "grid":
        grid_batch_size = int(batch_size**2 * n_items / n_users)
        sampler = GridSampler(
            dataset_shape=(n_users, n_items),
            approximate_batch_size=grid_batch_size,
            shuffle=shuffle,
        )
    elif sampler_type == "user":
        sampler = UserSampler(
            size=n_users,
            batch_size=batch_size,
            shuffle=shuffle,
        )
    elif sampler_type == "item":
        sampler = ItemSampler(
            size=n_items,
            batch_size=batch_size,
            shuffle=shuffle,
        )
    else:
        raise ValueError(f"Unknown sampler type {sampler_type}.")
    return sampler


def build_recommending_dataloader(
    dataset,
    sampler_type: Literal["grid", "user", "item"] = "user",
    batch_size=100,
    num_workers=4,
    persistent_workers=False,
    shuffle=False,
):
    sampler = build_recommending_sampler(
        batch_size=batch_size,
        n_users=dataset.shape[0],
        n_items=dataset.shape[1],
        sampler_type=sampler_type,
        shuffle=shuffle,
    )
    dataloader = DataLoader(
        dataset=dataset,
        sampler=sampler,
        batch_size=None,
        num_workers=num_workers,
        pin_memory=isinstance(num_workers, int) and num_workers > 1,
        persistent_workers=persistent_workers,
    )
    return dataloader


class SparseDataModuleInterface:
    def train_explicit(self) -> "csr_matrix" or None:
        return

    def val_explicit(self) -> "csr_matrix" or None:
        return

    def test_explicit(self) -> "csr_matrix" or None:
        return


class SparseDataModuleBase(SparseDataModuleInterface, SparseTensorUnpacker):
    def build_dataloader(self, **kwargs):
        return build_recommending_dataloader(**kwargs)

    def train_dataloader(self):
        if (explicit := self.train_explicit()) is not None:
            return self.build_dataloader(
                dataset=SparseDataset(explicit),
                sampler_type="user",
                shuffle=True,
            )

    def val_dataloader(self):
        if (explicit := self.val_explicit()) is not None:
            return self.build_dataloader(
                dataset=SparseDataset(explicit),
                sampler_type="user",
            )

    def test_dataloader(self):
        if (explicit := self.test_explicit()) is not None:
            return self.build_dataloader(
                dataset=SparseDataset(explicit),
                sampler_type="user",
            )



import einops
import numpy as np
import torch
import wandb

from .interface import RecommendingLossInterface
from .maths import kl_divergence, pairwise_difference


class MSELoss(RecommendingLossInterface):
    def __call__(
        self, explicit, model_ratings=None, model=None, user_ids=None, item_ids=None
    ):
        if model_ratings is None:
            model_ratings = model(user_ids=user_ids, item_ids=item_ids)
        explicit = explicit.to_dense()
        loss = ((model_ratings - explicit) ** 2).mean()
        return loss


class MSEL1Loss(RecommendingLossInterface):
    def __init__(self, explicit, ratings_deviation=1, mean_unobserved_rating=0):
        super().__init__(explicit)
        self.ratings_variance = ratings_deviation**2
        self.mean_unobserved_rating = mean_unobserved_rating

    def __call__(
        self, explicit, model_ratings=None, model=None, user_ids=None, item_ids=None
    ):
        if model_ratings is None:
            model_ratings = model(user_ids=user_ids, item_ids=item_ids)
        explicit = explicit.to_dense()
        observed_mask = explicit > 0
        loss = (
            observed_mask
            * (model_ratings - explicit) ** 2
            / (2 * self.ratings_variance)
        )
        loss += ~observed_mask * (model_ratings - self.mean_unobserved_rating).abs()
        return loss.sum()


class MSEProbabilityLoss(RecommendingLossInterface):
    """
    Recommender based on the following data generation model:
    1. Each user-item pair has a true, unknown, rating
    generated from normal distribution N_{r_ui, var}.
    2. Each rating has a probability to be observed p_ui = f(r_ui),
    where f is a non-decreasing function.
    That way higher quality items are more likely to be observed.

    Then the likelihood of given data sample is:
    likelihood(sample) =
    prod_observed_ui(p_ui * N_{r_ui, var}(rating_ui))) * prod_unobserved_ui(1 - p_ui)

    argmax(likelihood(sample)) =
    argmax(prod_observed_ui(f(r_ui) * N_{r_ui, var}(rating_ui))) * prod_unobserved_ui(1 - f(r_ui))) =
    argmax(sum_observed_ui(log f(r_ui) - (r_ui - rating_ui) ** 2 / (2 * var)) + sum_unobserved_ui(log(1 - f(r_ui))))

    The default for f is f(x) = c * x, where c is estimated as density / mean_rating
    """

    def __init__(self, explicit, ratings_deviation=1.0):
        super().__init__(explicit)
        self.ratings_variance = ratings_deviation**2
        mean_observed_rating = explicit.data.mean()
        density = len(explicit.data) / np.prod(explicit.shape)
        self.probability_to_be_observed_coef = density / mean_observed_rating

    def __call__(self, model, explicit, user_ids, item_ids):
        explicit = explicit.to_dense()
        observed_mask = explicit > 0
        ratings = model(user_ids=user_ids, item_ids=item_ids)
        loss = observed_mask * (ratings - explicit) ** 2 / (2 * self.ratings_variance)
        prob_to_be_observed = self.probability_to_be_observed_coef * ratings
        prob_to_be_observed = torch.clip(prob_to_be_observed, 1e-7, 1 - 1e-7)
        loss -= observed_mask * torch.log(prob_to_be_observed)
        loss -= ~observed_mask * torch.log(1 - prob_to_be_observed)
        return loss.sum()


def probability_of_user_preferences(
    ratings_of_supposedly_better_items, ratings_of_supposedly_worse_items
):
    """
    Given two rating matrices A, B of shape (n_users, n_items) corresponding
    to the same users, returns matrix C of shape (n_users, n_items, n_items)
    such that C[u, i, j] = Probability({user u prefers item A[u, i] over B[u, j]})
    """
    # Now P(item with rating 1 is more relevant than unrated item) =
    # = scaling_coefficient * sigmoid(1 - 0) = confidence_in_rating_quality
    probability = torch.sigmoid(
        pairwise_difference(
            ratings_of_supposedly_better_items, ratings_of_supposedly_worse_items
        )
    )
    return probability


class PersonalizedRankingLoss(RecommendingLossInterface):
    """Ranking based loss inspired by Bayesian Personalized Ranking paper."""

    def __init__(self, confidence=40):
        self.confidence = confidence

    def __call__(self, explicit, model_ratings):
        explicit = (explicit * (1 + self.confidence)).to_dense()
        model_ratings = model_ratings.to_dense()
        estimated_probs = torch.sigmoid(pairwise_difference(explicit, explicit))
        predicted_probs = torch.sigmoid(
            pairwise_difference(model_ratings, model_ratings)
        )
        # maybe in other order:
        # loss = kl_divergence(predicted_probs, estimated_probs)
        raise NotImplementedError
        # This is not a prob distribution!!!
        loss = kl_divergence(estimated_probs, predicted_probs)
        return loss / explicit.numel()

    # def __call__(self, *, explicit, model_ratings):
    #     explicit = explicit.coalesce()
    #     ids_of_users_who_rated_something = explicit.indices()[0].unique()
    #     if ids_of_users_who_rated_something.numel() == 0:
    #         return None
    #
    #     n_users, n_items = explicit.size()
    #     item_ids = torch.arange(n_items)
    #     items_mask = torch.full((n_items,), True)
    #     criterion = 0
    #     for i, user_id in enumerate(ids_of_users_who_rated_something):
    #         explicit_user_feedback = explicit[user_id]
    #         liked_item_ids = explicit_user_feedback.indices().squeeze()
    #         items_mask[:] = True
    #         items_mask[liked_item_ids] = False
    #         uninteracted_item_ids = item_ids[items_mask]
    #
    #         predicted_user_ratings = model_ratings[user_id]
    #
    #         user_ranking_correctness_criterion = -torch.log1p(
    #             torch.exp(
    #                 -(
    #                     predicted_user_ratings[liked_item_ids, None]
    #                     - predicted_user_ratings[uninteracted_item_ids]
    #                 )
    #             )
    #         ).mean()
    #         criterion += user_ranking_correctness_criterion
    #
    #     loss = -criterion / len(ids_of_users_who_rated_something)
    #     return loss


class PersonalizedRankingLossFast(RecommendingLossInterface):
    def __init__(self, max_rating=5, confidence_in_rating=0.9, memory_upper_bound=1e8):
        """
        :param max_rating: maximum value on the ratings scale
        :param confidence_in_rating: probability that item with best rating
        is more relevant than an unrated item
        :param memory_upper_bound: O(n) bound on size of tensor that can fit in the memory
        """
        self.scaling_coefficient = confidence_in_rating * (1 + torch.e**-max_rating)
        self.memory_upper_bound = memory_upper_bound

    def probability_of_user_preferences(
        self, ratings_of_supposedly_better_items, ratings_of_supposedly_worse_items
    ):
        """
        Given two rating matrices A, B of shape (n_users, n_items) corresponding
        to the same users, returns matrix C of shape (n_users, n_items, n_items)
        such that C[u, i, j] = Probability({user u prefers item A[u, i] over B[u, j]})
        """
        # Now P(item with rating 1 is more relevant than unrated item) =
        # = scaling_coefficient * sigmoid(1 - 0) = confidence_in_rating_quality
        probability = torch.sigmoid(
            pairwise_difference(
                ratings_of_supposedly_better_items, ratings_of_supposedly_worse_items
            )
        )
        return self.scaling_coefficient * probability

    @staticmethod
    def kl_divergence(
        probability_of_user_preferences, predicted_probability_of_user_preferences
    ):
        return -probability_of_user_preferences * torch.log(
            predicted_probability_of_user_preferences
        )

    def dense_cutout_kl_divergence(
        self, explicit_dense_cutout, model_ratings_dense_cutout
    ):
        probability_of_user_preferences = self.probability_of_user_preferences(
            explicit_dense_cutout, explicit_dense_cutout
        )
        predicted_probability_of_user_preferences = (
            self.probability_of_user_preferences(
                model_ratings_dense_cutout, model_ratings_dense_cutout
            )
        )
        kl_divergence = self.kl_divergence(
            probability_of_user_preferences,
            predicted_probability_of_user_preferences,
        )
        return kl_divergence

    def rated_unrated_kl_divergence(
        self,
        explicit_dense_cutout,
        model_ratings_dense_cutout,
        unrated_items_model_ratings,
    ):
        probability_of_user_preferences = self.probability_of_user_preferences(
            explicit_dense_cutout, torch.zeros_like(unrated_items_model_ratings)
        )
        predicted_probability_of_user_preferences = (
            self.probability_of_user_preferences(
                model_ratings_dense_cutout, unrated_items_model_ratings
            )
        )
        kl_divergence = self.kl_divergence(
            probability_of_user_preferences, predicted_probability_of_user_preferences
        )
        complementary_kl_divergence = self.kl_divergence(
            1 - probability_of_user_preferences,
            1 - predicted_probability_of_user_preferences,
        )
        return kl_divergence + complementary_kl_divergence

    def get_dense_and_unrated_splits(self, explicit):
        users_who_rated_something = explicit.indices()[0].unique()
        rated_items, counts = explicit.indices()[1].unique(return_counts=True)

        n_users, n_items = explicit.size()
        discriminant = n_items**2 - 4 * self.memory_upper_bound / n_users
        if discriminant > 0:
            n_rated_items_upper_bound = (n_items - discriminant**0.5) // 2
            if len(rated_items) > n_rated_items_upper_bound:
                n_clipped_items = len(rated_items) - n_rated_items_upper_bound
                wandb.log(
                    {"number of clipped items because of memory bound": n_clipped_items}
                )
                rated_items = rated_items[
                    counts.argsort(descending=True)[:n_rated_items_upper_bound]
                ]

        item_ids = torch.arange(n_items)
        items_mask = torch.full((n_items,), True)
        items_mask[rated_items] = False
        unrated_items = item_ids[items_mask]

        return users_who_rated_something, rated_items, unrated_items

    def split_explicit_feedback_and_ratings_into_dense_and_unrated(
        self, explicit, model_ratings
    ):
        (
            users_who_rated_something,
            rated_items,
            unrated_items,
        ) = self.get_dense_and_unrated_splits(explicit)

        explicit_dense_cutout = explicit.to_dense()[users_who_rated_something][
            :, rated_items
        ]
        model_ratings_dense_cutout = model_ratings[users_who_rated_something][
            :, rated_items
        ]
        unrated_items_model_ratings = model_ratings[users_who_rated_something][
            :, unrated_items
        ]
        return (
            explicit_dense_cutout,
            model_ratings_dense_cutout,
            unrated_items_model_ratings,
        )

    def __call__(self, *, explicit, model_ratings, implicit=None):
        explicit = explicit.to_sparse_coo().coalesce()
        if explicit.values().numel() == 0:
            return 0

        (
            explicit_dense_cutout,
            model_ratings_dense_cutout,
            unrated_items_model_ratings,
        ) = self.split_explicit_feedback_and_ratings_into_dense_and_unrated(
            explicit, model_ratings
        )

        dense_cutout_kl_divergence = self.dense_cutout_kl_divergence(
            explicit_dense_cutout, model_ratings_dense_cutout
        )

        rated_unrated_kl_divergence = self.rated_unrated_kl_divergence(
            explicit_dense_cutout,
            model_ratings_dense_cutout,
            unrated_items_model_ratings,
        )

        unrated_items_in_dense_cutout = explicit_dense_cutout != 0
        dense_cutout_mask = (
            unrated_items_in_dense_cutout[:, :, None]
            | unrated_items_in_dense_cutout[:, None, :]
        )
        dense_cutout_mask[
            :, (arange := torch.arange(dense_cutout_mask.shape[-1])), arange
        ] = False
        assert dense_cutout_mask.shape == dense_cutout_kl_divergence.shape
        loss = (dense_cutout_mask * dense_cutout_kl_divergence).sum()
        n_items_contributing_to_loss = dense_cutout_mask.sum()

        if n_unrated_items := unrated_items_model_ratings.shape[1]:
            rated_unrated_mask = einops.repeat(
                unrated_items_in_dense_cutout,
                f"user item -> user item {n_unrated_items}",
            )
            assert rated_unrated_mask.shape == rated_unrated_kl_divergence.shape
            loss += (rated_unrated_mask * rated_unrated_kl_divergence).sum()
            n_items_contributing_to_loss += rated_unrated_mask.sum()

        loss /= n_items_contributing_to_loss
        return loss



import torch
from torch.utils.data import DataLoader

from my_tools.models import register_regularization_hook
from my_tools.utils import to_torch_coo, torch_sparse_slice

from ..lit import LitRecommenderBase
from ..data import SparseDataset, build_recommending_dataloader, GridIterableDataset
from ..interface import RecommenderModuleBase
from ..utils import build_bias, build_weight, batch_size_in_bytes, Timer


class MatrixFactorization(RecommenderModuleBase):
    """Predicted_rating = user_factors @ item_factors, with bias and L2 regularization"""

    def __init__(self, latent_dimension=10, weight_decay=1.0e-3, **kwargs):
        super().__init__(**kwargs)
        self.weight_decay = weight_decay
        self.user_weight = build_weight(self.n_users, latent_dimension)
        self.user_bias = build_bias(self.n_users, 1)
        self.item_weight = build_weight(self.n_items, latent_dimension)
        self.item_bias = build_bias(self.n_items, 1)
        self.bias = build_bias(1)

    def forward(self, user_ids, item_ids):
        user_weight = self.user_weight[user_ids]
        user_bias = self.user_bias[user_ids]
        item_weight = self.item_weight[item_ids]
        item_bias = self.item_bias[item_ids]

        rating = user_weight @ item_weight.T + user_bias + item_bias.T + self.bias

        # Need to add regularization here because otherwise optimizer will decay all weights,
        # not only those corresponding to user and item ids. Also, it is important to add gradient
        # hooks after the forward calculations, otherwise decay messes with the model.
        for parameter in [user_weight, item_weight, user_bias, item_bias]:
            register_regularization_hook(parameter, self.weight_decay)

        return rating


class ConstrainedProbabilityMatrixFactorization(MatrixFactorization):
    def __init__(self, *args, latent_dimension=10, **kwargs):
        super().__init__(*args, latent_dimension=latent_dimension, **kwargs)
        self.item_rating_effect_weight = build_weight(self.n_items, latent_dimension)
        self.implicit_feedback_normalized = None

    def init_implicit_feedback_normalized(self):
        implicit_feedback = self.to_scipy_coo(self.explicit) > 0
        implicit_feedback_normalized = implicit_feedback.multiply(
            1 / (implicit_feedback.sum(axis=1) + 1e-8)
        ).astype("float32")
        self.implicit_feedback_normalized = to_torch_coo(implicit_feedback_normalized)

    def forward(self, user_ids, item_ids):
        if self.implicit_feedback_normalized is None:
            self.init_implicit_feedback_normalized()

        # Need to clone to avoid gradient hook accumulation on same tensor and subsequent memory leak
        item_rating_effect_weight = self.item_rating_effect_weight.clone()

        item_weight = self.item_weight[item_ids]

        users_implicit_feedback = torch_sparse_slice(
            self.implicit_feedback_normalized, row_ids=user_ids
        ).to(self.bias.device)
        user_weights_offset_caused_by_their_ratings = (
            users_implicit_feedback @ item_rating_effect_weight
        )

        ratings = super().forward(user_ids=user_ids, item_ids=item_ids)
        ratings += user_weights_offset_caused_by_their_ratings @ item_weight.T

        # Scale down regularization because item_rating_effect_weight is decayed
        # for each batch, whereas other parameters have only their slices decayed.
        scale_down = self.user_weight.shape[0] / len(user_ids)
        register_regularization_hook(
            item_rating_effect_weight, self.weight_decay / scale_down
        )
        return ratings


class MFSlimRecommender(RecommenderModuleBase):
    """
    y = x @ w, w = (1 - I) * (encoder @ decoder)

    Zeroing out the diagonal to prevent model from fitting to
    the identity matrix and enforcing the model to predict
    item's relevance based on other items. Slim-like models
    are nice because they work online, are interpretable,
    simple and scalable.
    """

    def __init__(
        self,
        latent_dimension=10,
        l2_regularization=0.0,
        l1_regularization=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.encoder = build_weight(self.n_items, latent_dimension)
        self.decoder = build_weight(latent_dimension, self.n_items)
        self.item_bias = build_bias(self.n_items)
        self.bias = build_bias(1)
        self.l2_regularization = l2_regularization
        self.l1_regularization = l1_regularization

    @Timer()
    def online_ratings(self, users_explicit):
        users_explicit = users_explicit.to(self.device, torch.float32)
        encoder = self.encoder.clone()
        decoder = self.decoder.clone()
        diag = torch.einsum("id, di -> i", encoder, decoder)
        ratings = (
            users_explicit @ encoder @ decoder
            - users_explicit.to_dense() * diag
            + self.item_bias
            + self.bias
        )
        if self.l2_regularization > 0 or self.l1_regularization > 0:
            for parameter in [encoder, decoder]:
                register_regularization_hook(
                    tensor=parameter,
                    l2_coefficient=self.l2_regularization,
                    l1_coefficient=self.l1_regularization,
                )
        return ratings


class MFRecommender(LitRecommenderBase):
    @property
    def class_candidates(self):
        return super().class_candidates + [
            MatrixFactorization,
            ConstrainedProbabilityMatrixFactorization,
            MFSlimRecommender,
        ]

    def train_dataloader(self):
        config = self.hparams["datamodule"]
        n_items = self.hparams["n_items"]
        n_users = self.hparams["n_users"]
        batch_size = config.get("batch_size", 100)
        grid_batch_size = int(batch_size**2 * n_items / n_users)
        dataset = GridIterableDataset(
            dataset_shape=(n_users, n_items),
            approximate_batch_size=grid_batch_size,
            shuffle=True,
        )
        num_workers = config.get("num_workers", 0)
        dataloader = DataLoader(
            dataset=dataset,
            batch_size=None,
            num_workers=num_workers,
            persistent_workers=config.get("persistent_workers", False),
            pin_memory=isinstance(num_workers, int) and num_workers > 1,
        )
        return dataloader

    def val_dataloader(self):
        config = self.hparams["datamodule"]
        batch_size = config.get(
            "val_batch_size", self.hparams["datamodule"].get("batch_size", 100)
        )
        return build_recommending_dataloader(
            dataset=SparseDataset(self.val_explicit()),
            sampler_type="user",
            batch_size=batch_size,
            num_workers=config.get("num_workers", 0),
            persistent_workers=config.get("persistent_workers", False),
        )

    def test_dataloader(self):
        config = self.hparams["datamodule"]
        batch_size = config.get(
            "test_batch_size", self.hparams["datamodule"].get("batch_size", 100)
        )
        return build_recommending_dataloader(
            dataset=SparseDataset(self.test_explicit()),
            sampler_type="user",
            batch_size=batch_size,
            num_workers=config.get("num_workers", 0),
            persistent_workers=config.get("persistent_workers", False),
        )

    def training_step(self, batch, batch_idx):
        self.log("train_batch_size_in_bytes", float(batch_size_in_bytes(batch)))
        explicit = torch_sparse_slice(
            sparse_matrix=self.model.explicit,
            row_ids=batch["user_ids"],
            col_ids=batch["item_ids"],
        ).to(self.device)
        loss = self.loss(model=self.model, explicit=explicit, **batch)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        self.log("val_batch_size_in_bytes", float(batch_size_in_bytes(batch)))
        loss = self.loss(model=self.model, **batch)
        self.log("val_loss", loss)
        return loss



from typing import Literal, TYPE_CHECKING

import pandas as pd
import torch
import wandb

from my_tools.utils import BuilderMixin
from .data import MovieLens25m, MovieLens100k, csv_imdb_ratings_to_dataframe
from ..callbacks import RecommendingExplanationCallback
from ..lit import LitRecommenderBase

if TYPE_CHECKING:
    from .data import MovieLensInterface
    from ..interface import RecommenderModuleBase


class RecommendingExplanationIMDBCallback(
    RecommendingExplanationCallback, BuilderMixin
):
    def __init__(
        self,
        path_to_imdb_ratings_csv: str,
        n_recommendations=10,
        ratings_scale_max_to_convert_to=5,
    ):
        super().__init__(n_recommendations=n_recommendations)
        self.path_to_imdb_ratings_csv = path_to_imdb_ratings_csv
        self.ratings_scale_max_to_convert_to = ratings_scale_max_to_convert_to
        self.imdb_ratings: pd.DataFrame or None = None
        self.movielens: MovieLensInterface or None = None

    def setup(self, trainer, pl_module: LitRecommenderBase, stage):
        if self.imdb_ratings is None:
            self.imdb_ratings = csv_imdb_ratings_to_dataframe(
                path_to_imdb_ratings_csv=self.path_to_imdb_ratings_csv,
                ratings_scale_max_to_convert_to=self.ratings_scale_max_to_convert_to,
            )
            self.movielens = pl_module.movielens
            self.users_explicit = self.movielens.imdb_ratings_dataframe_to_explicit(
                imdb_ratings=self.imdb_ratings
            )

    @property
    def class_candidates(self):
        return [MovieLens100k, MovieLens25m]

    def on_train_start(self, trainer, pl_module):
        wandb.log(dict(imdb_ratings=wandb.Table(dataframe=self.imdb_ratings)))
        _, indices = torch.topk(
            self.users_explicit.to_dense()[0], k=self.n_recommendations
        )
        dataframe = self.movielens.items_description(dense_item_ids=indices.numpy())
        title = f"sanity_check/imdb_as_recommendations"
        wandb.log({title: wandb.Table(dataframe=dataframe)})

    def my_on_epoch_end(
        self,
        model: "RecommenderModuleBase",
        stage: 'Literal["train", "val", "test", "predict"]',
    ):
        super().my_on_epoch_end(model=model, stage=stage)
        recommendations = model.online_recommend(
            users_explicit=self.users_explicit,
            n_recommendations=self.n_recommendations,
        )
        recommendations = recommendations.cpu().numpy()[0]
        items_description = self.movielens.items_description(
            dense_item_ids=recommendations
        )
        title = f"{stage} recommendations based on imdb ratings"
        wandb.log({title: wandb.Table(dataframe=items_description)})

        try:
            online_ratings = model.online_ratings(users_explicit=self.users_explicit)
        except NotImplementedError:
            pass
        else:
            _, indices = torch.topk(online_ratings[0], k=self.n_recommendations)
            dataframe = self.movielens.items_description(
                dense_item_ids=indices.cpu().numpy()
            )
            title = f"sanity_check/{stage} unfiltered recommendations"
            wandb.log({title: wandb.Table(dataframe=dataframe)})



from typing import TYPE_CHECKING

import pytorch_lightning as pl
import torch

from my_tools.utils import BuilderMixin

from . import losses, models
from .data import build_recommending_dataloader, SparseDataModuleBase
from .utils import prepare_artifacts_from_config

if TYPE_CHECKING:
    from .interface import RecommenderModuleBase, RecommendingLossInterface


class LitRecommenderBase(SparseDataModuleBase, pl.LightningModule, BuilderMixin):
    def __init__(self, n_users=None, n_items=None, **config):
        super().__init__()
        self.save_hyperparameters()
        prepare_artifacts_from_config(config=config)
        if n_users is None or n_items is None:
            n_users, n_items = self.train_explicit().shape
            self.save_hyperparameters()  # This will update n_users, n_items.
        self.model: "RecommenderModuleBase" = self.build_model()
        self.loss: "RecommendingLossInterface" or None = None
        if "loss" in config:
            self.loss = self.build_class(
                explicit=self.train_explicit(), **config["loss"]
            )

    def build_model(self):
        model_config = self.hparams["model"]
        return self.build_class(
            explicit=self.train_explicit(),
            n_users=self.hparams["n_users"],
            n_items=self.hparams["n_items"],
            **model_config,
        )

    @property
    def module_candidates(self):
        return [models, losses, torch.optim, torch.optim.lr_scheduler]

    @property
    def class_candidates(self):
        return [LRLambda]

    def build_dataloader(self, **kwargs):
        config = self.hparams["datamodule"]
        return build_recommending_dataloader(
            batch_size=config.get("batch_size", 100),
            num_workers=config.get("num_workers", 0),
            persistent_workers=config.get("persistent_workers", False),
            **kwargs,
        )

    def configure_optimizers(self):
        optimizer_config = self.hparams["optimizer"].copy()
        optimizer = self.build_class(params=self.parameters(), **optimizer_config)
        if (lr_scheduler_config := self.hparams.get("lr_scheduler")) is None:
            return optimizer

        if lr_scheduler_config["class_name"] == "LambdaLR":
            lr_scheduler_config = lr_scheduler_config.copy()
            lr_lambda_config = lr_scheduler_config.pop["lr_lambda"]
            lr_scheduler_config["lr_lambda"] = self.build_class(**lr_lambda_config)

        lr_scheduler = self.build_class(optimizer=optimizer, **lr_scheduler_config)
        return dict(optimizer=optimizer, lr_scheduler=lr_scheduler)

    def forward(self, **batch):
        return self.model(user_ids=batch["user_id"], item_ids=batch["item_id"])

    """Step placeholders to enable hooks without error."""

    def training_step(self, batch, batch_idx):
        pass

    def validation_step(self, batch, batch_idx):
        pass

    def test_step(self, batch, batch_idx):
        pass


class LRLambda:
    """Learning rate scheduler with linear warmup and subsequent reverse square root fading."""

    def __init__(self, warmup_epochs=20):
        self.warmup_epochs = warmup_epochs

    def __call__(self, epoch):
        epoch += 1
        learning_rate_multiplier = min(
            (self.warmup_epochs / epoch) ** 0.5, epoch / self.warmup_epochs
        )
        return learning_rate_multiplier


class NonGradientRecommenderMixin:
    def on_train_batch_start(self: LitRecommenderBase, batch, batch_idx):
        if self.current_epoch == 0:
            if batch_idx == 0:
                self.model.fit()
        else:
            self.trainer.should_stop = True
            return -1

    def configure_optimizers(self: pl.LightningModule):
        return



import pandas as pd

from ..models.cat import CatboostAggregatorFromArtifacts, CatboostRecommenderBase
from .data import MovieLens100k, MovieLens25m


class CatboostMovieLens100kFeatureRecommender(CatboostRecommenderBase):
    def __init__(self, *, movielens_directory, **kwargs):
        super().__init__(**kwargs)
        self.movielens = MovieLens100k(movielens_directory)

    def user_features(self):
        user_features = self.movielens["u.user"].reset_index(names="user_id")
        user_features["user_id"] = self.movielens.dataset_to_dense_user_ids(
            dataset_user_ids=user_features["user_id"].values
        )
        return self.FeatureReturnValue(
            dataframe=user_features,
            cat_features={"user_id", "gender", "occupation", "zip code"},
        )

    def item_features(self):
        item_features = (
            self.movielens["u.item"]
            .reset_index(names="item_id")
            .drop(["movie title", "video release date", "IMDb URL"], axis="columns")
        )
        item_features["release date"] = pd.to_datetime(
            item_features["release date"]
        ).astype(int)
        return self.FeatureReturnValue(
            dataframe=item_features,
            cat_features=set(item_features.columns.drop(["release date"])),
        )

    def user_item_features(self):
        user_item_features = self.movielens["u.data"][
            ["user_id", "item_id", "timestamp"]
        ]
        return self.FeatureReturnValue(dataframe=user_item_features)


class CatboostMovieLens100kFeatureAggregatorFromArtifacts(
    CatboostAggregatorFromArtifacts, CatboostMovieLens100kFeatureRecommender
):
    pass


class CatboostMovieLens25mFeatureRecommender(CatboostRecommenderBase):
    def __init__(self, *, movielens_directory, **kwargs):
        super().__init__(**kwargs)
        self.movielens = MovieLens25m(movielens_directory)

    def item_features(self):
        item_features = self.movielens["movies"].reset_index(names="item_id")
        item_features = item_features.query(
            "item_id in @rated_item_ids",
            local_dict=dict(rated_item_ids=self.movielens.unique_dataset_item_ids),
        )
        genres = item_features["genres"].str.get_dummies()
        item_features = item_features.drop(["title", "genres"], axis="columns")
        item_features = item_features.join(genres)
        item_features["item_id"] = self.movielens.dataset_to_dense_item_ids(
            dataset_item_ids=item_features["item_id"].values
        )
        return self.FeatureReturnValue(
            dataframe=item_features,
            cat_features=set(item_features.columns),
        )

    def user_item_features(self):
        user_item_features = self.movielens["ratings"][
            ["user_id", "item_id", "timestamp"]
        ]
        user_item_features = self.set_index(
            user_item_features, kind=self.FeatureKind.user_item
        )
        user_item_tags = (
            self.movielens["tags"]
            .groupby(["user_id", "item_id"])["tag"]
            .apply(", ".join)
        )
        user_item_features = user_item_features.join(user_item_tags)
        return self.FeatureReturnValue(
            dataframe=user_item_features,
            text_features={"tag"},
        )


class CatboostMovieLens25mFeatureAggregatorFromArtifacts(
    CatboostAggregatorFromArtifacts, CatboostMovieLens25mFeatureRecommender
):
    pass



from my_tools.entrypoint import LightningConfigBuilder, ConfigDispenser

from . import callbacks, metrics, models
from .movielens import lit as movielens_lit
from .movielens import callbacks as movielens_callbacks
from .utils import wandb_context_manager, init_torch


class RecommendingBuilder(LightningConfigBuilder):
    @property
    def module_candidates(self):
        return super().module_candidates + [
            callbacks,
            metrics,
            models,
            movielens_callbacks,
            movielens_lit,
        ]

    def build_lightning_module(self):
        class_name = self.config["lightning_module"]["class_name"]
        lightning_module = self.build_class(class_name=class_name, **self.config)
        return lightning_module


class RecommendingConfigDispenser(ConfigDispenser):
    def from_cli(self):
        parser = self.cli_argument_parser
        parser.add_argument(
            "--cross-validate",
            "-cv",
            action="store_true",
            help="whether to run cross validation",
        )
        parser.add_argument(
            "--wandb-save",
            "-s",
            action="store_true",
            help="whether to run save wandb checkpoint artifact",
        )
        cli_args = vars(parser.parse_args())
        self.__init__(**cli_args)


def fit(config):
    with wandb_context_manager(config):
        init_torch(debug=True)
        constructor = RecommendingBuilder(config)
        lightning_module = constructor.build_lightning_module()
        trainer = constructor.build_trainer()
        trainer.fit(lightning_module)
        trainer.test(lightning_module)
    return lightning_module



import asyncio
import functools
import os.path
import time
from typing import TYPE_CHECKING, TextIO

from aenum import Enum, MultiValue, Unique
import gradio as gr
import numpy as np
import requests
import scipy
import torch
import wandb

from .movielens.data import csv_imdb_ratings_to_dataframe
from .utils import Timer

if TYPE_CHECKING:
    from .interface import RecommenderModuleBase
    from .movielens.data import MovieLensInterface


class Feedback(Enum, settings=(MultiValue, Unique), init="str rating"):
    not_interested = "Haven't seen it and not interested", 0
    interested = "Haven't seen it, but I'm interested", 3.5
    one = "⭐️", 1
    two = "⭐️⭐️", 2
    three = "⭐️⭐️⭐️", 3
    four = "⭐️⭐️⭐️⭐️", 4
    five = "⭐️⭐️⭐️⭐️⭐️", 5


def load_model(torch_nn_module, state_dict_path):
    state_dict = torch.load(state_dict_path)
    model = torch_nn_module()
    model.load_state_dict(state_dict)
    return model


class MovieMarkdownGenerator:
    def __init__(self, movielens, tmdb_api_token):
        self.movielens = movielens
        self.tmdb_api_token = tmdb_api_token

    @staticmethod
    def imdb_url(imdb_id):
        return f"https://www.imdb.com/title/tt{imdb_id:07d}"

    @staticmethod
    def tmdb_request_url(imdb_id, tmdb_api_token):
        return (
            f"https://api.themoviedb.org/3/find/tt{imdb_id:07d}"
            f"?api_key={tmdb_api_token}&external_source=imdb_id"
        )

    @staticmethod
    def poster_url_from_tmdb_poster_path(poster_path, size="w200"):
        return f"https://image.tmdb.org/t/p/{size}{poster_path}"

    @staticmethod
    @functools.lru_cache(maxsize=10_000)
    def cached_request(url):
        """Caching requests saves a lot of time."""
        return requests.get(url)

    @Timer()
    def __call__(self, item_id):
        movielens_id = self.movielens.dense_to_dataset_item_ids([item_id])[0]
        imdb_id = self.movielens.dataset_to_imdb_movie_ids([movielens_id])[0]
        tmdb_request_url = self.tmdb_request_url(
            imdb_id=imdb_id, tmdb_api_token=self.tmdb_api_token
        )
        request = self.cached_request(url=tmdb_request_url)
        if not request.ok:
            wandb.log(
                dict(
                    failed_request=dict(
                        reason=request.reason,
                        status_code=request.status_code,
                        text=request.text,
                    )
                )
            )
            markdown = self.movielens["movies"].loc[movielens_id]["movie title"]
            return markdown, ""

        movie_results = request.json()["movie_results"][0]
        poster_url = self.poster_url_from_tmdb_poster_path(
            poster_path=movie_results["poster_path"]
        )
        title = movie_results["title"]
        imdb_url = self.imdb_url(imdb_id=imdb_id)
        markdown = f"""[{title}]({imdb_url})"""
        return markdown, poster_url


class Session:
    def __init__(
        self,
        recommender: "RecommenderModuleBase",
        movie_markdown_generator: MovieMarkdownGenerator,
        n_recommendations=10,
    ):
        self.recommender = recommender
        self.movie_markdown_generator = movie_markdown_generator
        self.movielens: "MovieLensInterface" = self.movie_markdown_generator.movielens
        self.n_recommendations = n_recommendations

        wandb.define_metric("session_length", summary="max")
        self.session_length = 0
        self.session_id = time.time()
        self.explicit = np.full(self.recommender.n_items, fill_value=np.nan)
        self.movie_id, self.initial_markdown, self.initial_poster_url = self.content()
        self.content_task = self.content()

    def __deepcopy__(self, memo):
        """
        Gradio makes a copy of state block at some point, I don't know why exactly.
        The problem is that you can't copy a task, and returning new instance instead
        doesn't break anything, so that's what I'm going to do.
        """
        return Session(
            recommender=self.recommender,
            movie_markdown_generator=self.movie_markdown_generator,
        )

    def content(self):
        explicit = scipy.sparse.coo_matrix(np.nan_to_num(self.explicit).reshape(1, -1))
        with torch.no_grad():
            user_recommendations = self.recommender.online_recommend(
                users_explicit=explicit, n_recommendations=self.recommender.n_items
            )[0].numpy()

        new_items_mask = np.isnan(self.explicit)
        new_items_positions = new_items_mask[user_recommendations].nonzero()[0]
        movie_id = user_recommendations[new_items_positions[0]]

        # To not recommend same item twice before rating gets saved.
        self.explicit[movie_id] = 0

        markdown, poster_url = self.movie_markdown_generator(item_id=movie_id)

        return movie_id, markdown, poster_url

    async def async_content(self):
        return self.content()

    def save_rating(self, rating, movie_id):
        if self.explicit[movie_id] != 0:
            raise ValueError(
                f"Movie {movie_id} was already rated {self.explicit[movie_id]}."
            )
        self.explicit[movie_id] = rating
        self.session_length += 1
        wandb.log(
            dict(
                session_id=self.session_id,
                movie_id=movie_id,
                rating=rating,
                session_length=self.session_length,
            )
        )

    async def next_content(self, feedback: Feedback):
        if asyncio.isfuture(self.content_task):
            with Timer(name="await_content"):
                content = await self.content_task
        else:
            content = self.content_task
        next_movie_id, markdown, poster_url = content
        self.save_rating(rating=feedback.rating, movie_id=self.movie_id)
        wandb.log(
            dict(
                interested=float(feedback == Feedback.interested),
                not_interested=float(feedback == Feedback.not_interested),
            )
        )
        self.movie_id = next_movie_id
        self.content_task = asyncio.create_task(self.async_content())
        return markdown, poster_url

    def upload(self, file_object: TextIO) -> str:
        """Function for file block upload method."""
        filename = file_object.name
        imdb_ratings = csv_imdb_ratings_to_dataframe(path_to_imdb_ratings_csv=filename)
        try:
            explicit = self.movielens.imdb_ratings_dataframe_to_explicit(
                imdb_ratings=imdb_ratings
            )
        except Exception as e:
            wandb.log(dict(upload_error=str(e)))
            return f"Couldn't read imdb ratings file: \n{str(e)}"

        with torch.no_grad():
            recommendations = self.recommender.online_recommend(
                explicit, n_recommendations=self.n_recommendations
            )

        markdown = ""
        for i, item_id in enumerate(recommendations[0].numpy(), start=1):
            movie_markdown, poster_url = self.movie_markdown_generator(item_id=item_id)
            markdown += f"""{i}. {movie_markdown} ![poster]({poster_url}) \n"""
        return markdown


def interactive_feedback_click(feedback: Feedback):
    async def click_inner(session: Session):
        markdown, poster_url = await session.next_content(feedback=feedback)
        return session, markdown, poster_url

    return click_inner


def gradio_relative_html_path(relative_path):
    return os.path.join("file", relative_path)


def build_app_blocks(
    recommender: "RecommenderModuleBase",
    movie_markdown_generator: MovieMarkdownGenerator,
    media_directory: str,
    n_recommendations_from_imdb_ratings=10,
) -> None:

    gr.Markdown("This movie recommender adapts to your preferences")
    session_state = gr.State(
        value=Session(
            recommender=recommender,
            movie_markdown_generator=movie_markdown_generator,
            n_recommendations=n_recommendations_from_imdb_ratings,
        )
    )

    with gr.Tab("Interactively rate movies"):
        with gr.Row():
            image_block = gr.Image(
                session_state.value.initial_poster_url, interactive=False
            )
            image_block.style(width=200)
            with gr.Column(scale=4):
                recommendations_markdown_block = gr.Markdown(
                    session_state.value.initial_markdown
                )

                def build_button(feedback: Feedback):
                    button = gr.Button(feedback.str)
                    button.click(
                        fn=interactive_feedback_click(feedback),
                        inputs=[session_state],
                        outputs=[
                            session_state,
                            recommendations_markdown_block,
                            image_block,
                        ],
                    )
                    return button

                with gr.Row():
                    build_button(Feedback.not_interested)
                    build_button(Feedback.interested)
                with gr.Row():
                    for rated_feedback in list(Feedback)[2:]:
                        build_button(rated_feedback)

    with gr.Tab("Upload IMDb ratings"):

        def image_html(filename, width=400):
            path = gradio_relative_html_path(os.path.join(media_directory, filename))
            return f"""<img src="{path}" width="{width}">"""

        howto_markdown = gr.Markdown(
            f"""
            If you have a profile on IMDb, you can upload the ratings from your profile page:
            
            1. Go to [imdb.com](https://www.imdb.com)
            2. Click on your profile and then on the "Your ratings" tab
            {image_html("step1.png")}
            3. Click on the ellipsis menu on "Your ratings" page
            {image_html("step2.png")}
            4. Select "Export"
            {image_html("step3.png")}
            5. Now your ratings will be downloaded in csv format, and you can upload them below
            """
        )
        file_block = gr.File(interactive=True)
        recommendations_markdown_block = gr.Markdown("")
        file_block.upload(
            fn=lambda file: (
                howto_markdown.update(visible=False),
                session_state.value.upload(file),
            ),
            inputs=[file_block],
            outputs=[howto_markdown, recommendations_markdown_block],
        )



import abc
import functools
import os
import pickle
import sys
import warnings
from collections import Counter
from dataclasses import dataclass, field
from enum import Enum
from typing import TYPE_CHECKING, Tuple, List, Set

import catboost
import einops
import numpy as np
import pandas as pd
import shap
import torch
import wandb

from scipy.sparse import coo_array
from tqdm.auto import tqdm

from my_tools.utils import torch_sparse_slice

from ..interface import (
    ExplanationMixin,
    FitExplicitInterfaceMixin,
    RecommenderModuleBase,
)
from ..movielens import lit as movielens_lit
from ..utils import (
    save_shap_force_plot,
    wandb_plt_figure,
    Timer,
    profile,
    clipped_tensor_string_representation,
)

if TYPE_CHECKING:
    from my_tools.utils import SparseTensor


def batch_fit_catboost_ranker(
    cb_params: dict,
    pool: catboost.Pool,
    n_groups_per_batch: int,
    cb_fit_verbose: int = 100,
    disable_tqdm: bool = False,
) -> catboost.CatBoostRanker:
    """Fits catboost ranker with cb_params to pool in batches with n non-homogenous groups per batch."""
    catboost_ranker = catboost.CatBoostRanker(**cb_params)
    if pool.shape[0] == 0:
        return catboost_ranker
    _, group_indices = np.unique(pool.get_group_id_hash(), return_index=True)
    begins = range(0, len(group_indices), n_groups_per_batch)
    ends = list(begins)[1:] + [len(group_indices)]

    if n_groups_per_batch == -1 or len(begins) <= n_groups_per_batch:
        catboost_ranker.fit(pool, verbose=cb_fit_verbose)
        return catboost_ranker

    if pool.get_text_feature_indices():
        raise ValueError(
            "Models summation is not supported for models with text features, "
            "so either pass n_groups_per_batch=-1, or remove text features."
        )

    batch_rankers = []
    if not disable_tqdm:
        print(
            f"Fitting catboost on "
            f"{'gpu' if cb_params.get('task_type') == 'GPU' else 'cpu'}"
            f" to pool of shape {pool.shape}.",
            file=sys.stderr,
        )
    for begin, end in tqdm(
        iterable=zip(begins, ends), total=len(begins), disable=disable_tqdm
    ):
        batch_indices = np.arange(group_indices[begin], group_indices[end])
        batch_pool = pool.slice(batch_indices)
        if catboost_ranker.is_fitted():
            batch_pool.set_baseline(baseline=catboost_ranker.predict(batch_pool))
        batch_ranker = catboost.CatBoostRanker(**cb_params)
        batch_ranker.fit(batch_pool, verbose=cb_fit_verbose)
        batch_rankers.append(batch_ranker)

    catboost_ranker = catboost.sum_models(batch_rankers)
    return catboost_ranker


def split_by_group_sizes(
    group_ids: np.array, group_size_threshold: int
) -> "Tuple[np.array, np.array]":
    """
    Returns an array with indices of items that have
    group_ids of groups with no more than
    group_size_threshold members, and the rest indices.
    """
    group_ids = pd.DataFrame(dict(group_id=group_ids))
    less_group_ids = (
        group_ids.groupby("group_id")
        .size()
        .rename("size")
        .to_frame()
        .query(f"size <= {group_size_threshold}")
        .index
    )
    less_indices = group_ids.query(
        "group_id in @less_group_ids", local_dict=dict(less_group_ids=less_group_ids)
    ).index.values
    more_indices = group_ids.index.difference(less_indices).values
    return less_indices, more_indices


def fit_catboost_ranker(
    pool: catboost.Pool,
    cb_params: dict,
    cpu_n_users_per_batch=10_000,
    gpu_n_users_per_batch=1000,
    max_gpu_query_size=1023,
) -> catboost.CatBoostRanker:
    """
    There seems to be a bug in catboost code with gpu training that leads to inferior results.

    While this function is a great candidate for async - gpu and cpu
    fitting are pretty independent, but creating coroutines in any
    function requires for it to be async itself, and any function that calls
    it, and so forth, propagating up to the entry point. This makes async
    code non-modular, fragile and bug prone.
    """
    if cb_params.get("task_type") != "GPU":
        return batch_fit_catboost_ranker(
            cb_params=cb_params,
            pool=pool,
            n_groups_per_batch=cpu_n_users_per_batch,
        )

    gpu_indices, cpu_indices = split_by_group_sizes(
        group_ids=pool.get_group_id_hash(), group_size_threshold=max_gpu_query_size
    )
    gpu_ranker = batch_fit_catboost_ranker(
        cb_params=cb_params,
        pool=pool.slice(gpu_indices),
        n_groups_per_batch=gpu_n_users_per_batch,
    )
    cpu_ranker = batch_fit_catboost_ranker(
        cb_params={**cb_params, "task_type": "CPU"},
        pool=pool.slice(cpu_indices),
        n_groups_per_batch=cpu_n_users_per_batch,
    )
    catboost_ranker = catboost.sum_models(
        [gpu_ranker, cpu_ranker], weights=[len(gpu_indices), len(cpu_indices)]
    )
    return catboost_ranker


class CatboostInterface(
    RecommenderModuleBase, FitExplicitInterfaceMixin, ExplanationMixin, abc.ABC
):
    class FeatureKind(Enum):
        user = "user"
        item = "item"
        user_item = "user_item"

        @property
        def index_columns(self):
            return {
                self.user: ["user_id"],
                self.item: ["item_id"],
                self.user_item: ["user_id", "item_id"],
            }[self]

    @dataclass
    class FeatureReturnValue:
        dataframe: pd.DataFrame or None = None
        cat_features: "Set[str]" = field(default_factory=set)
        text_features: "Set[str]" = field(default_factory=set)

    def __init__(
        self,
        cb_params=None,
        cb_cpu_fit_n_users_per_batch=10_000,
        cb_gpu_fit_n_users_per_batch=1000,
        cb_fit_verbose=100,
        use_user_ids_as_features=True,
        use_item_ids_as_features=True,
        use_text_features=True,
        unknown_token="__NA__",
        feature_perturbation="interventional",
        max_gpu_query_size=1023,
        **kwargs,
    ):
        """There seems to be a bug in catboost code with gpu training that leads to inferior results."""
        super().__init__(**kwargs)
        self._cb_params = cb_params
        self.cb_cpu_fit_n_users_per_batch = cb_cpu_fit_n_users_per_batch
        self.cb_gpu_fit_n_users_per_batch = cb_gpu_fit_n_users_per_batch
        self.cb_fit_verbose = cb_fit_verbose
        self.use_user_ids_as_features = use_user_ids_as_features
        self.use_item_ids_as_features = use_item_ids_as_features
        self.use_text_features = use_text_features
        self.unknown_token = unknown_token
        self.feature_perturbation = feature_perturbation
        self.max_gpu_query_size = max_gpu_query_size

        if self.cb_params.get("task_type") == "GPU" and "devices" in self.cb_params:
            os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
            os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(
                map(str, self.cb_params["devices"])
            )
        self.model = catboost.CatBoostRanker(**self.cb_params)
        self.cat_features = {i: set() for i in self.FeatureKind}
        self.text_features = {i: set() for i in self.FeatureKind}
        self._train_user_item_rating_dataframe = None
        self._shap_explainer = None
        warnings.simplefilter(action="ignore", category=FutureWarning)

    @property
    def cb_params(self) -> dict:
        """Convenience property to avoid unwanted inplace changes to self._cb_params."""
        return self._cb_params.copy() if self._cb_params else {}

    @property
    def shap_explainer(self):
        if self._shap_explainer is None:
            if not self.model.is_fitted():
                raise ValueError(
                    "To initialize shap.TreeExplainer, model must first be fitted."
                )
            self._shap_explainer = shap.TreeExplainer(
                self.model, feature_perturbation=self.feature_perturbation
            )
        return self._shap_explainer

    def get_extra_state(self):
        self.model.save_model("tmp")
        with open("tmp", "rb") as f:
            catboost_bytes = f.read()
        shap_explainer_bytes = pickle.dumps(self._shap_explainer)
        return dict(
            catboost_bytes=catboost_bytes, shap_explainer_bytes=shap_explainer_bytes
        )

    def set_extra_state(self, bytes_dict):
        with open("tmp", "wb") as f:
            f.write(bytes_dict["catboost_bytes"])
        self.model.load_model("tmp")
        self._shap_explainer = pickle.loads(bytes_dict["shap_explainer_bytes"])

    def user_features(self) -> FeatureReturnValue:
        """
        Returns FeatureReturnValue with these kwargs:
            dataframe: dataset specific user features with at least a "user_id" column
            cat_features: set of column names with categorical features
            text_features: set of column names with text features
        """
        return self.FeatureReturnValue()

    def item_features(self) -> FeatureReturnValue:
        """Returns dict analogous to user_features, but with "item_id" column."""
        return self.FeatureReturnValue()

    def user_item_features(self) -> FeatureReturnValue:
        """Returns dict analogous to user_features, but with "user_id" and "item_id" columns."""
        return self.FeatureReturnValue()

    @staticmethod
    def set_index(dataframe: pd.DataFrame, kind: FeatureKind):
        if dataframe.index.names != kind.index_columns:
            dataframe = dataframe.set_index(kind.index_columns, verify_integrity=True)
        if not dataframe.index.is_unique:
            raise ValueError(f"DataFrame index is not unique:\n{dataframe.head()}")
        return dataframe.sort_index()

    @functools.lru_cache()
    def cached_indexed_features_dataframe(self, kind: FeatureKind) -> pd.DataFrame:
        features: CatboostInterface.FeatureReturnValue = getattr(
            self, f"{kind.value}_features"
        )()
        self.cat_features[kind] |= features.cat_features
        self.text_features[kind] |= features.text_features
        if features.dataframe is not None:
            features_dataframe = self.set_index(features.dataframe, kind=kind)
            return features_dataframe

    @Timer()
    def merge_features(self, user_item_dataframe):
        """Merges dataset specific features defined in *_features methods into the given user-item dataframe."""
        for kind in self.FeatureKind:
            features_dataframe = self.cached_indexed_features_dataframe(kind=kind)
            if features_dataframe is not None:
                user_item_dataframe = user_item_dataframe.join(features_dataframe)
        return user_item_dataframe

    def rating_stats(self, ratings_dataframe, kind: FeatureKind) -> pd.DataFrame:
        """Returns dataframe with kind.index_columns index or columns and some ratings statistics in columns."""
        return pd.DataFrame(
            columns=CatboostInterface.FeatureKind.user_item.index_columns
        )

    @functools.lru_cache()
    def cached_rating_stats(self, kind: FeatureKind) -> pd.DataFrame:
        ratings_dataframe = self.ratings_user_item_dataframe(explicit=self.explicit)
        rating_stats = self.rating_stats(ratings_dataframe=ratings_dataframe, kind=kind)
        rating_stats = self.set_index(rating_stats, kind=kind)
        return rating_stats

    def merge_rating_stats(
        self, dataframe, users_explicit=None, user_ids: np.array = None
    ) -> pd.DataFrame:
        """
        Merges user and item stats from rating_stats method into the
        received user-item dataframe, using cached stats from train explicit matrix,
        and recalculating only if the online user stats are needed.
        """
        dataframe = self.set_index(dataframe, kind=self.FeatureKind.user_item)
        user_ratings_stats = self.cached_rating_stats(kind=self.FeatureKind.user)
        item_ratings_stats = self.cached_rating_stats(kind=self.FeatureKind.item)
        if user_ids is not None and self.are_user_ids_fictive(user_ids):
            ratings_dataframe = self.ratings_user_item_dataframe(
                explicit=users_explicit, user_ids=user_ids
            )
            user_ratings_stats = self.rating_stats(
                ratings_dataframe=ratings_dataframe, kind=self.FeatureKind.user
            )
        dataframe = dataframe.join(user_ratings_stats).join(item_ratings_stats)
        return dataframe

    def ratings_user_item_dataframe(
        self, explicit: "SparseTensor", user_ids: np.array = None
    ) -> pd.DataFrame:
        """
        Builds a dataframe with user_id, item_id multiindex and rating
        column from user-item pairs present in the explicit feedback matrix.
        """
        assert explicit.shape[1] == self.n_items, "Invalid explicit matrix shape."
        explicit = self.to_scipy_coo(explicit)
        user_ids = np.arange(self.n_users) if user_ids is None else user_ids
        dataframe = pd.DataFrame(
            dict(
                user_id=user_ids[explicit.row],
                item_id=explicit.col,
                rating=explicit.data,
            )
        )
        dataframe = self.set_index(dataframe=dataframe, kind=self.FeatureKind.user_item)
        return dataframe

    def extract_pool_kwargs(self, dataframe) -> dict:
        """
        Builds dict of catboost Pool kwargs from dataframe with at least
        user_id, item_id, *cat_features, *text_features,
        some number of numeric features columns,
        and rating column if it is a labeled dataset.
        """
        dataframe = dataframe.reset_index()
        group_id = dataframe["user_id"].values
        cat_features = set().union(*self.cat_features.values())
        text_features = set().union(*self.text_features.values())
        if "rating" in dataframe:
            label = dataframe["rating"].values
            dataframe = dataframe.drop("rating", axis="columns")
        else:
            label = None
        if self.use_user_ids_as_features:
            cat_features |= {"user_id"}
        else:
            dataframe = dataframe.drop("user_id", axis="columns")
            cat_features -= {"user_id"}
        if self.use_item_ids_as_features:
            cat_features |= {"item_id"}
        else:
            dataframe = dataframe.drop("item_id", axis="columns")
            cat_features -= {"item_id"}
        if not self.use_text_features:
            dataframe = dataframe.drop(text_features, axis="columns")
            text_features = []
        return dict(
            dataframe=dataframe,
            group_id=group_id,
            label=label,
            cat_features=list(cat_features),
            text_features=list(text_features),
        )

    def postprocess_pool_kwargs(
        self,
        dataframe: pd.DataFrame,
        group_id: np.array,
        label: np.array or None = None,
        cat_features: List[str] = (),
        text_features: List[str] = (),
    ) -> dict:
        """
        Only minor datatype manipulations are performed in this method
        to comply with what catboost expects, also missing categories
        and text features are marked with self.unknown_token.
        """
        dataframe = dataframe.copy()
        for column in dataframe:
            if column in list(cat_features):
                dataframe[column] = (
                    dataframe[column]
                    .fillna(self.unknown_token)
                    .astype("str")
                    .astype("category")
                )
            elif column in list(text_features):
                dataframe[column] = (
                    dataframe[column].fillna(self.unknown_token).astype("string")
                )
            else:
                dataframe[column] = pd.to_numeric(dataframe[column])
        return dict(
            data=dataframe,
            cat_features=cat_features,
            text_features=text_features,
            label=label,
            group_id=group_id,
        )

    def build_labeled_user_item_dataframe(self, explicit) -> pd.DataFrame:
        """
        Given explicit feedback matrix, return dataframe with user-item pairs
        present in the explicit matrix, plus any extra features, except for dataset
        specific ones, which will be added automatically later in the pipeline.
        """
        dataframe = self.ratings_user_item_dataframe(explicit=explicit)
        dataframe = self.merge_rating_stats(dataframe=dataframe)
        return dataframe

    @property
    @functools.lru_cache()
    def train_user_item_dataframe(self):
        """Returns the dataframe that the model fits to."""
        return self.build_labeled_user_item_dataframe(explicit=self.explicit)

    @abc.abstractmethod
    def build_recommend_user_item_dataframe(
        self, users_explicit: "SparseTensor", user_ids: np.array, n_recommendations: int
    ) -> pd.DataFrame:
        """
        Given users' feedback matrix, their ids and requested number of
        recommendations, build a dataframe with user-item pairs,
        where user ids are the ones from received user_ids, and item_ids
        are the possible candidates for catboost ranker to rank.
        """
        ...

    def pool(self, user_item_dataframe):
        """
        Convenience method that consolidates the
        catboost Pool building pipeline from dataframe received from
        either build_labeled_dataframe or build_recommend_dataframe method.
        """
        user_item_dataframe = self.merge_features(user_item_dataframe)
        pool_kwargs = self.extract_pool_kwargs(dataframe=user_item_dataframe)
        pool_kwargs = self.postprocess_pool_kwargs(**pool_kwargs)
        pool = catboost.Pool(**pool_kwargs)
        return pool

    def fit(self):
        pool = self.pool(self.train_user_item_dataframe)
        self.model = fit_catboost_ranker(
            pool=pool,
            cb_params=self.cb_params,
            cpu_n_users_per_batch=self.cb_cpu_fit_n_users_per_batch,
            gpu_n_users_per_batch=self.cb_gpu_fit_n_users_per_batch,
            max_gpu_query_size=self.max_gpu_query_size,
        )

    def forward(self, user_ids, item_ids):
        ...
        raise NotImplementedError(
            "Catboost works well when recommending small number of items, "
            "and is not meant to generate ratings for all items."
        )

    def flat_catboost_predicted_ratings_to_matrix(
        self, user_ids: np.array, item_ids: np.array, flat_predicted_ratings: np.array
    ) -> "SparseTensor":
        """Unravel the user-item rating pairs into a matrix by mapping the coordinates."""
        row_ids = user_ids.astype(np.int32)
        col_ids = item_ids.astype(np.int32)
        ratings = coo_array(
            (flat_predicted_ratings, (row_ids, col_ids)),
            shape=[row_ids.max() + 1, self.n_items],
        )
        ratings = ratings.tocsr()[np.unique(row_ids)]
        ratings = self.to_torch_coo(ratings).to(torch.float32).to_dense()
        ratings[ratings == 0] = -torch.inf
        return ratings

    def common_recommend(
        self,
        users_explicit: "SparseTensor",
        user_ids,
        n_recommendations: int or None,
    ):
        n_recommendations = n_recommendations or self.n_items
        dataframe = self.build_recommend_user_item_dataframe(
            users_explicit=users_explicit,
            user_ids=user_ids,
            n_recommendations=n_recommendations,
        )
        pool = self.pool(dataframe)
        flat_predicted_ratings = self.model.predict(pool)
        ratings = self.flat_catboost_predicted_ratings_to_matrix(
            user_ids=dataframe.index.get_level_values("user_id").values,
            item_ids=dataframe.index.get_level_values("item_id").values,
            flat_predicted_ratings=flat_predicted_ratings,
        )
        recommendations = self.ratings_to_filtered_recommendations(
            explicit=users_explicit,
            ratings=ratings,
            n_recommendations=n_recommendations,
        )
        return recommendations

    @Timer()
    def recommend(self, user_ids, n_recommendations=None):
        return self.common_recommend(
            users_explicit=torch_sparse_slice(self.explicit, row_ids=user_ids),
            user_ids=user_ids,
            n_recommendations=n_recommendations,
        )

    def fictive_user_ids(self, n_users) -> torch.IntTensor:
        """
        Returns fictive user ids to use as group ids during online recommend,
        which do not intersect with user ids the model saw during fit.
        """
        return torch.arange(self.n_users, self.n_users + n_users)

    def are_user_ids_fictive(self, user_ids: np.array) -> bool:
        """
        Predicate to determine if user ids are fictive, or,
        equivalently, if we are performing an online recommend.
        """
        return any(user_ids >= self.n_users)

    @Timer()
    def online_recommend(self, users_explicit, n_recommendations=None):
        return self.common_recommend(
            users_explicit=users_explicit,
            user_ids=self.fictive_user_ids(n_users=users_explicit.shape[0]),
            n_recommendations=n_recommendations,
        )

    def get_feature_importance(self, pool) -> pd.DataFrame:
        """
        Given unlabeled pool, returns the default importance type
        for each feature, which is usually the average feature's
        impact on the prediction value or loss function.
        """
        less_indices, more_indices = split_by_group_sizes(
            group_ids=pool.get_group_id_hash(),
            group_size_threshold=self.max_gpu_query_size,
        )
        pool = pool.slice(less_indices)
        return self.model.get_feature_importance(pool, prettified=True)

    def shap_kwargs(self, user_item_dataframe):
        """
        Given user_item_dataframe, returned by one of the build_*_user_item_dataframe methods,
        returns all the necessary parameters to start drawing graphs using the shap module.
        """
        user_item_dataframe = self.merge_features(
            user_item_dataframe=user_item_dataframe
        )
        pool_kwargs = self.extract_pool_kwargs(dataframe=user_item_dataframe)
        pool = catboost.Pool(**self.postprocess_pool_kwargs(**pool_kwargs))
        shap_values = self.shap_explainer.shap_values(pool)
        return dict(
            base_value=self.shap_explainer.expected_value,
            shap_values=shap_values,
            features=pool_kwargs["dataframe"],
        )

    @Timer()
    def explain_recommendations_for_user(
        self,
        user_id=None,
        user_explicit=None,
        n_recommendations=10,
        log=False,
        logging_prefix="explanation/",
        plot_waterfalls=False,
        figsize=(8, 5),
    ) -> dict:
        return_explanations_dict = {}
        if user_id is not None:
            user_ids = torch.IntTensor([user_id])
            recommendations = self.recommend(
                user_ids=user_ids, n_recommendations=n_recommendations
            )
            user_explicit = torch_sparse_slice(self.explicit, row_ids=user_ids)
        else:
            recommendations = self.online_recommend(
                users_explicit=user_explicit, n_recommendations=n_recommendations
            )
            user_ids = self.fictive_user_ids(n_users=1)

        recommendations = recommendations[0].cpu().numpy()
        dataframe = self.build_recommend_user_item_dataframe(
            users_explicit=user_explicit,
            user_ids=user_ids,
            n_recommendations=n_recommendations,
        )
        dataframe = dataframe.query(
            "user_id in @user_ids.numpy() and item_id in @recommendations"
        )
        shap_kwargs = self.shap_kwargs(user_item_dataframe=dataframe)
        return_explanations_dict["features"] = shap_kwargs["features"]
        if log:
            title = logging_prefix + "features"
            wandb.log({title: wandb.Table(dataframe=shap_kwargs["features"])})

        shap_plot = shap.force_plot(
            **shap_kwargs,
            out_names="Predicted relevance of items for users",
            text_rotation=0,
        )
        force_plot_textio = save_shap_force_plot(shap_plot=shap_plot)
        return_explanations_dict["force_plot_textio"] = force_plot_textio
        if log:
            title = logging_prefix + "force plot"
            wandb.log({title: wandb.Html(force_plot_textio)})

        with wandb_plt_figure(
            title=logging_prefix + "decision plot", figsize=figsize, log=log
        ) as decision_plot_figure:
            shap.decision_plot(
                **shap_kwargs, legend_labels=[f"item {i}" for i in recommendations]
            )
        return_explanations_dict["decision_plot_figure"] = decision_plot_figure

        if plot_waterfalls:
            return_explanations_dict["waterfall_figures"] = {}
            dataframe = self.merge_features(dataframe)
            postprocessed_pool_kwargs = self.postprocess_pool_kwargs(
                **self.extract_pool_kwargs(dataframe=dataframe)
            )
            shap_explanation: shap.Explanation = self.shap_explainer(
                X=postprocessed_pool_kwargs["data"]
            )
            for item_id, item_explanation in zip(recommendations, shap_explanation):
                with wandb_plt_figure(
                    title=logging_prefix + f"waterfall plot for item {item_id}",
                    figsize=figsize,
                    log=log,
                ) as figure:
                    shap.waterfall_plot(item_explanation)
                return_explanations_dict["waterfall_figures"][item_id] = figure

        return return_explanations_dict


class CatboostRecommenderBase(CatboostInterface):
    def rating_stats(
        self, ratings_dataframe, kind: CatboostInterface.FeatureKind
    ) -> pd.DataFrame:
        """Returns dataframe with kind.index_columns index and some ratings statistics in columns."""
        if kind not in [
            self.FeatureKind.user,
            self.FeatureKind.item,
        ]:
            raise ValueError(f"Rating features are undefined for kind {kind}.")
        kind = kind.value
        mean_ratings = (
            ratings_dataframe.groupby(f"{kind}_id")["rating"]
            .mean()
            .rename(f"mean_{kind}_ratings")
        )
        n_ratings = (
            ratings_dataframe.groupby(f"{kind}_id").size().rename(f"{kind}_n_ratings")
        )
        dataframe = pd.concat([mean_ratings, n_ratings], axis="columns")
        return dataframe

    def build_recommend_user_item_dataframe(
        self, users_explicit, user_ids, n_recommendations
    ):
        dataframe = pd.DataFrame(
            dict(
                user_id=np.repeat(user_ids.numpy(), self.n_items),
                item_id=np.tile(np.arange(self.n_items), len(user_ids)),
            )
        )
        dataframe = self.merge_rating_stats(
            dataframe=dataframe, users_explicit=users_explicit, user_ids=user_ids
        )
        return dataframe


class CatboostAggregatorRecommender(CatboostInterface):
    """This recommender uses other pre-fit models' topk recommendations as features."""

    def __init__(
        self,
        fit_recommenders: "list[RecommenderModuleBase]" = None,
        recommender_names: "list[str]" = None,
        train_n_recommendations=10,
        dataframe_build_batch_size=10_000,
        **kwargs,
    ):
        if (
            fit_recommenders is not None
            and recommender_names is not None
            and len(fit_recommenders) != len(recommender_names)
        ):
            raise ValueError("Recommenders and their names must be of equal length.")
        super().__init__(**kwargs)
        self.fit_recommenders: "torch.nn.ModuleList[RecommenderModuleBase]" = (
            torch.nn.ModuleList(fit_recommenders)
        )
        self.recommender_names = recommender_names or self.generate_model_names(
            self.fit_recommenders
        )
        self.train_n_recommendations = train_n_recommendations
        self.dataframe_build_batch_size = dataframe_build_batch_size

    @staticmethod
    def generate_model_names(models):
        names = []
        counter = Counter()
        for model in models:
            name = model.__class__.__name__
            names.append(f"{name}_{counter[name]}")
            counter.update([name])
        return names

    # def user_item_features(self):
    #     return self.FeatureReturnValue(cat_features=set(self.recommender_names))

    def ranks_dataframe(self, user_ids, item_ids, ranks):
        """
        :param user_ids: tensor of shape [n_rows]
        :param item_ids: tensor of shape [n_rows]
        :param ranks: tensor of shape [n_rows, len(self.fit_recommenders)]
        :return: dataframe with self.recommender_names columns,
            user_id and item_id multiindex
        """
        dataframe = pd.DataFrame(
            data=torch.cat(
                [user_ids.reshape(-1, 1), item_ids.reshape(-1, 1), ranks], dim=1
            ),
            columns=["user_id", "item_id"] + self.recommender_names,
        )
        dataframe = self.set_index(dataframe, kind=self.FeatureKind.user_item)
        return dataframe

    @Timer()
    @profile(log_every=10)
    def poll_fit_recommenders(
        self, users_explicit, user_ids
    ) -> "List[torch.IntTensor]":
        """
        Given users' explicit feedback and their user_ids, returns
        a list of length equal to len(self.fit_recommenders), with
        torch matrices with the ranks of items that each separate fit_recommender
        assigned for each user, if the item was recommended at all,
        and -1 otherwise.
        """
        extended_ranks = 1 + torch.arange(self.n_items + 1).repeat(len(user_ids), 1)
        extended_ranks[:, -1] = -1
        recommenders_ranks = []
        for recommender, name in zip(self.fit_recommenders, self.recommender_names):
            if self.are_user_ids_fictive(user_ids.numpy()):
                item_ids = recommender.online_recommend(users_explicit)
            else:
                item_ids = recommender.recommend(user_ids)

            item_ids = torch.where(
                torch.BoolTensor(item_ids == recommender.invalid_recommendation_mark),
                self.n_items,
                item_ids,
            )
            recommender_ranks = torch.scatter(
                input=-torch.ones_like(extended_ranks),
                dim=1,
                index=item_ids,
                src=extended_ranks,
            )[:, :-1]
            recommenders_ranks.append(recommender_ranks)

        return recommenders_ranks

    @Timer()
    def aggregate_topk_recommendations(
        self, users_explicit, user_ids, n_recommendations
    ):
        """
        Method to determine item recommendation candidates for each user during recommending.
        Returns user-item dataframe with self.recommender_names columns
        induced by aggregated topk n_recommendations by self.fit_recommenders for
        users with corresponding user_ids. Values in self.recommender_names
        columns at index [user_id, item_id] are equal to rank of item_id in recommender's
        recommendations for user_id.
        """
        recommenders_ranks = self.poll_fit_recommenders(
            user_ids=user_ids, users_explicit=users_explicit
        )
        items_reciprocal_ranks_sum = sum(
            [torch.where(i == -1, 0, 1 / i) for i in recommenders_ranks]
        )
        _, topk_item_ids = items_reciprocal_ranks_sum.topk(k=n_recommendations, dim=1)
        topk_ranks = [
            torch.take_along_dim(input=recommender_rank, indices=topk_item_ids, dim=1)
            for recommender_rank in recommenders_ranks
        ]
        repeated_user_ids = einops.repeat(user_ids, f"u -> u {n_recommendations}")
        topk_item_ids = einops.rearrange(topk_item_ids, "u k -> (u k)")
        topk_ranks = einops.rearrange(topk_ranks, "r u k -> (u k) r")
        dataframe = self.ranks_dataframe(
            user_ids=repeated_user_ids, item_ids=topk_item_ids, ranks=topk_ranks
        )
        return dataframe

    def user_batches(
        self, user_ids: torch.IntTensor, users_explicit: "SparseTensor", tqdm_stage=None
    ):
        """Batched iterator over synchronized user_ids and users_explicit."""
        for batch_indices in tqdm(
            iterable=torch.arange(len(user_ids)).split(
                split_size=self.dataframe_build_batch_size
            ),
            desc=f"Building {tqdm_stage or ''} dataframe for users: {clipped_tensor_string_representation(user_ids)}",
            disable=not bool(tqdm_stage),
        ):
            batch_user_ids = user_ids[batch_indices]
            batch_users_explicit = torch_sparse_slice(
                users_explicit, row_ids=batch_indices
            )
            yield batch_user_ids, batch_users_explicit

    def build_labeled_user_item_dataframe(self, explicit):
        if explicit.shape != (
            expected_shape := torch.Size([self.n_users, self.n_items])
        ):
            raise ValueError(
                "Shape of explicit feedback matrix doesn't "
                "match the expected shape of [self.n_users, self.n_items]:"
                f"{explicit.shape} != {expected_shape}"
            )

        dataframes = []
        for batch_user_ids, batch_users_explicit in self.user_batches(
            user_ids=torch.arange(self.n_users),
            users_explicit=explicit,
            tqdm_stage="labeled",
        ):
            recommenders_ranks = self.poll_fit_recommenders(
                user_ids=batch_user_ids,
                users_explicit=batch_users_explicit,
            )
            user_pos, item_ids = (
                self.to_torch_coo(batch_users_explicit).coalesce().indices()
            )
            train_ranks = [rank[user_pos, item_ids] for rank in recommenders_ranks]
            train_ranks = einops.rearrange(train_ranks, "r n -> n r")
            train_dataframe = self.ranks_dataframe(
                user_ids=batch_user_ids[user_pos], item_ids=item_ids, ranks=train_ranks
            )
            dataframes.append(train_dataframe)

        dataframe = pd.concat(dataframes)
        super_dataframe = super().build_labeled_user_item_dataframe(explicit=explicit)
        assert (
            len(dataframe.index.symmetric_difference(super_dataframe.index)) == 0
        ), "Index of batch-constructed dataframe doesn't match the expected one."
        dataframe = dataframe.join(super_dataframe)
        return dataframe

    def build_recommend_user_item_dataframe(
        self, users_explicit, user_ids, n_recommendations
    ):
        dataframes = []
        for batch_user_ids, batch_users_explicit in self.user_batches(
            user_ids=user_ids, users_explicit=users_explicit
        ):
            dataframes.append(
                self.aggregate_topk_recommendations(
                    users_explicit=batch_users_explicit,
                    user_ids=batch_user_ids,
                    n_recommendations=n_recommendations,
                )
            )
        dataframe = pd.concat(dataframes)
        dataframe = self.merge_rating_stats(
            dataframe=dataframe, users_explicit=users_explicit, user_ids=user_ids
        )
        return dataframe


class CatboostAggregatorFromArtifacts(CatboostAggregatorRecommender):
    """
    Wrapper for CatboostAggregatorRecommender to fetch fit recommenders from wandb artifacts.
    To use this module, BuilderMixin *_candidates methods need to be overridden first.
    """

    def __init__(
        self,
        *args,
        recommender_artifacts: "list[str]" = (),
        lit_modules: "list[str]" = (),
        **kwargs,
    ):
        fit_recommenders = []
        for artifact_name, lit_name in (
            tqdm_progress_bar := tqdm(
                iterable=zip(recommender_artifacts, lit_modules),
                total=len(recommender_artifacts),
            )
        ):
            tqdm_progress_bar.set_description(f"Fetching artifact {artifact_name}")
            artifact = wandb.use_artifact(artifact_name)
            checkpoint_path = artifact.file()
            lit_module = getattr(movielens_lit, lit_name)
            lit_module = lit_module.load_from_checkpoint(checkpoint_path)
            recommender = lit_module.model
            fit_recommenders.append(recommender)
        super().__init__(
            *args,
            fit_recommenders=fit_recommenders,
            recommender_names=recommender_artifacts,
            **kwargs,
        )



from .als import ALS, ALSjit, ALSjitBiased
from .baseline import (
    RandomRecommender,
    PopularRecommender,
    SVDRecommender,
    ImplicitNearestNeighborsRecommender,
    ImplicitMatrixFactorizationRecommender,
    UnpopularSVDRecommender,
)
from .mf import (
    MatrixFactorization,
    ConstrainedProbabilityMatrixFactorization,
    MFSlimRecommender,
    MFRecommender,
)
from .slim import SLIM, SLIMDataset, SLIMRecommender



import abc
import collections
import contextlib
import functools
import inspect
import io
import os
import re
import sys
import time
import types
import warnings
from types import TracebackType, FunctionType
from typing import TextIO, Callable, Type, Dict, Generator

import line_profiler
import shap
import torch
import wandb
import yaml
from matplotlib import pyplot as plt

from torch.utils.data import random_split


def build_weight(*dimensions):
    weight = torch.nn.Parameter(torch.empty(*dimensions))
    torch.nn.init.xavier_normal_(weight)
    return weight


def build_bias(*dimensions):
    bias = torch.nn.Parameter(torch.zeros(*dimensions))
    return bias


def split_dataset(dataset, fraction):
    len_dataset = len(dataset)
    right_size = int(fraction * len_dataset)
    left_size = len_dataset - right_size
    return random_split(dataset, [left_size, right_size])


def fetch_artifact(
    *, entity=None, project=None, artifact_name, alias="latest", api_key=None
):
    wandb_api = wandb.Api(api_key=api_key)
    entity = entity or wandb.run.entity
    project = project or wandb.run.project
    artifact = wandb_api.artifact(f"{entity}/{project}/{artifact_name}:{alias}")
    return artifact


def load_path_from_artifact(artifact, path_inside_artifact="checkpoint"):
    artifact_dir = artifact.download()
    checkpoint_path = os.path.join(
        artifact_dir, artifact.get_path(path_inside_artifact).path
    )
    return checkpoint_path


def update_from_base_config(config, base_config_file):
    """Keeps everything from config, and updates it with entries from base config up to 1 depth in."""
    base_config = yaml.safe_load(open(base_config_file))
    for k, v in config.items():
        if k in base_config and isinstance(v, dict):
            base_config[k].update(v)
        else:
            base_config[k] = v
    return base_config


def wandb_context_manager(config):
    if wandb.run is None and config.get("logger") is not None:
        return wandb.init(project=config.get("project"), config=config)
    return contextlib.nullcontext()


@contextlib.contextmanager
def plt_figure(*args, title=None, **kwargs):
    figure = plt.figure(*args, **kwargs)
    if title is not None:
        plt.title(title)
    try:
        yield figure
    finally:
        plt.close(figure)


@contextlib.contextmanager
def wandb_plt_figure(title, log: bool = True, *args, **kwargs):
    with plt_figure(*args, title=title, **kwargs) as figure:
        yield figure
        if log:
            wandb.log({title: wandb.Image(figure)})


def tensor_size_in_bytes(tensor):
    return tensor.numel() * tensor.element_size()


def batch_size_in_bytes(batch):
    if torch.is_tensor(batch):
        return tensor_size_in_bytes(batch)
    elif isinstance(batch, dict):
        return sum(batch_size_in_bytes(i) for i in batch.values())
    return 0


@contextlib.contextmanager
def filter_warnings(action: str, category=None):
    with warnings.catch_warnings():
        warnings.simplefilter(action=action, category=category)
        yield


def save_shap_force_plot(shap_plot: shap.plots._force.BaseVisualizer) -> TextIO:
    textio = io.TextIOWrapper(io.BytesIO())
    shap.save_html(textio, shap_plot)
    textio.seek(0)
    return textio


class Singleton(type):
    """
    class A(metaclass=Singleton):
        pass

    a1 = A()
    a2 = A()
    a1 is a2
    True
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Logger(metaclass=Singleton):
    profilers_dict = collections.defaultdict(
        lambda: dict(profiler=line_profiler.LineProfiler(), accumulated_logs=0)
    )

    def log(self, dict_to_log: dict) -> None:
        raise NotImplementedError

    def log_timer(self, timer: "Timer") -> None:
        raise NotImplementedError

    def register_function(self, function: FunctionType) -> None:
        if function not in self.profilers_dict:
            profiler = self.profilers_dict[function]["profiler"]
            profiler.add_function(function)
            profiler.enable()

    def register_function_call(self, function, log_every=1) -> None:
        profiler_dict = self.profilers_dict[function]
        profiler_dict["accumulated_logs"] += 1
        if profiler_dict["accumulated_logs"] >= log_every:
            profiler_dict["accumulated_logs"] = 0
            self.log_profiled_function(
                profiler=profiler_dict["profiler"], function=function
            )

    def log_profiled_function(
        self, profiler: line_profiler.LineProfiler, function: FunctionType
    ) -> None:
        raise NotImplementedError


class FileLogger(Logger):
    def __init__(self, file: io.TextIOBase = sys.stderr):
        self.file = file

    # TODO: color
    def log(self, dict_to_log: dict):
        print(*[f"{k}:\t{v}" for k, v in dict_to_log.items()], sep="\n", file=self.file)

    def log_timer(self, timer):
        print(f"{timer.get_name()}:\t{timer.delta}", file=self.file)

    def log_profiled_function(
        self, profiler: line_profiler.LineProfiler, function: FunctionType
    ):
        profiler.print_stats(stream=self.file)


class WandbLogger(Logger):
    total_times = collections.defaultdict(lambda: 0)

    def log(self, dict_to_log: dict):
        if wandb.run is not None:
            wandb.log(dict_to_log)

    def log_timer(self, timer: "Timer"):
        name = timer.get_name()
        time_name = f"time/{name}"
        total_time_name = f"time/total/{name}"
        if wandb.run is not None:
            wandb.define_metric(time_name, summary="mean")
            wandb.define_metric(total_time_name, summary="last")
        self.total_times[name] += timer.delta
        self.log({time_name: timer.delta, total_time_name: self.total_times[name]})

    def log_profiled_function(
        self, profiler: line_profiler.LineProfiler, function: FunctionType
    ):
        stream = io.StringIO()
        profiler.print_stats(stream=stream)
        stream.seek(0)
        title = f"profile/{function.__qualname__}"
        wandb.log({title: wandb.Html(f"<pre>{stream.read()}</pre>")})


class DecoratorInterface:
    function: Callable = None

    def on_enter(self) -> None:
        """On enter hook."""

    def on_exit(self) -> None:
        """On exit hook."""

    def on_before_function_call(self, *args, **kwargs) -> None:
        """A hook, args and kwargs are the ones that were passed to the function."""

    def on_after_function_call(self, result, *args, **kwargs) -> None:
        """A hook, function returned result when it was called with args and kwargs."""

    def decorate(self, function: Callable) -> Callable:
        @functools.wraps(function)
        def wrap(*args, **kwargs):
            self.function = function
            self.on_enter()
            self.on_before_function_call(*args, **kwargs)
            result = function(*args, **kwargs)
            self.on_after_function_call(result, *args, **kwargs)
            self.on_exit()
            self.function = None
            return result

        return wrap

    __call__ = decorate


class Timer(DecoratorInterface):
    """Context manager and decorator for timing."""

    def __init__(self, name: str = "", logger: Logger = WandbLogger()):
        self.name = name
        self.logger = logger
        self.begin: float or None = None
        self.end: float or None = None

    def get_name(self):
        return self.name or (
            self.function.__qualname__ if self.function is not None else ""
        )

    @property
    def delta(self) -> float or None:
        return None if self.end is None else self.end - self.begin

    def on_enter(self):
        self.begin = time.time()
        self.end = None

    def on_exit(self):
        self.end = time.time()
        self.logger.log_timer(self)

    def __enter__(self):
        self.on_enter()
        return self

    def __exit__(
        self,
        exc_type: "Type[BaseException]" or None = None,
        exc_val: BaseException or None = None,
        exc_tb: TracebackType or None = None,
    ):
        self.on_exit()


class PerIterTimer(Timer):
    n_iters = None

    def set_n_iters(self, n_iters):
        self.n_iters = n_iters

    def on_exit(self):
        super().on_exit()
        if self.n_iters is not None:
            self.logger.log(
                {f"time/{self.get_name()} per iter": self.delta / self.n_iters}
            )


class MethodDecorator(DecoratorInterface, abc.ABC):
    """
    Class intended as base for class method decorators
    to provide the decorator with the instance whose
    method is being called and with the name of the method.
    """

    instance = None
    owner_class: type or None = None
    method_name: str or None = None

    def __call__(self, unbound_method: Callable) -> Callable:
        """Temporarily replace the unbound_method with self."""
        self.function = unbound_method
        return self

    def __set_name__(self, owner_class: type, method_name: str):
        """
        This method is called when owning class owner_class is created,
        here I replace the original unbound method with a wrapper
        that will be given an instance – self, do something with it in self.decorate."""
        self.owner_class = owner_class
        self.method_name = method_name

        @functools.wraps(self.function)
        def method(instance, *args, **kwargs):
            self.instance = instance
            function = self.decorate(
                function=functools.partial(self.function, instance)
            )
            return function(*args, **kwargs)

        setattr(owner_class, method_name, method)


class MethodsDecoratorClassDecorator:
    """Decorator for classes to decorate all user defined methods with decorator."""

    def __init__(self, decorator: Callable):
        self.decorator = decorator

    @staticmethod
    def is_protected_or_builtin_name(attribute_name: str) -> bool:
        return attribute_name.startswith("_")

    @staticmethod
    def is_function(obj) -> bool:
        return isinstance(obj, FunctionType)

    @staticmethod
    def is_functions_first_parameter_self(function: FunctionType) -> bool:
        signature = inspect.signature(function)
        return next(iter(signature.parameters), None) == "self"

    @staticmethod
    def is_unbound_method_defined_in_class(
        unbound_method: FunctionType, cls: type
    ) -> bool:
        return unbound_method.__qualname__.startswith(cls.__name__ + ".")

    def __call__(self, cls):
        for key in dir(cls):
            if self.is_protected_or_builtin_name(key):
                continue
            value = getattr(cls, key)
            if (
                self.is_function(obj=value)
                and self.is_functions_first_parameter_self(function=value)
                and self.is_unbound_method_defined_in_class(
                    unbound_method=value, cls=cls
                )
            ):
                setattr(cls, key, self.decorator(value))
        return cls


class TimerClassDecorator(MethodsDecoratorClassDecorator):
    def __init__(self):
        super().__init__(decorator=Timer())


def split_full_artifact_name(full_artifact_name: str) -> Dict[str, str]:
    g = lambda name: f"(?P<{name}>.+)"
    pattern = f"{g('entity')}/{g('project')}/{g('name')}:{g('alias')}"
    match = re.fullmatch(pattern=pattern, string=full_artifact_name)
    if match is None:
        raise ValueError(
            f"Full artifact name '{full_artifact_name}' did not match the expected pattern {pattern}."
        )
    return match.groupdict()


def prepare_artifacts_from_config(config: dict) -> None:
    for artifact_config in config.get("artifacts", []):
        prepare_artifact(**artifact_config)


def prepare_artifact(
    full_artifact_name: str,
    directory: str,
    artifact_type: str = None,
    update_with_local_directory: bool = False,
    match_directory_exactly: bool = False,
) -> wandb.Artifact:
    """
    :param full_artifact_name: artifact name in format entity/project/artifact_name:alias
    :param directory: directory to download to or update artifact from
    :param artifact_type: type of artifact
    :param update_with_local_directory: if true, then will create or update artifact using directory
    :param match_directory_exactly: if true, then content of directory will be checksummed against the
        artifact manifest and redownloaded if it doesn't exactly match the expected content.
        WARNING: This will DELETE all files in directory that are not included in the artifact.
    :return: prepared wandb artifact
    """
    if directory == ".":
        raise ValueError("Using cwd as artifact directory is bad practice.")
    if update_with_local_directory and (
        not os.path.exists(directory) or not os.listdir(directory)
    ):
        raise ValueError(
            f"Passing update_with_local_directory=True requires that directory {directory} exists and is not empty."
        )
    artifact_dict = split_full_artifact_name(full_artifact_name)
    if update_with_local_directory and (
        wandb.run.entity != artifact_dict["entity"]
        or wandb.run.project != artifact_dict["project"]
    ):
        raise ValueError(
            "Cannot update artifact with entity or project different from current ones."
        )

    if update_with_local_directory:
        artifact = wandb.log_artifact(
            artifact_or_path=directory,
            name=artifact_dict["name"],
            type=artifact_type,
            aliases=[artifact_dict["alias"]],
        )
    else:
        artifact = wandb.run.use_artifact(full_artifact_name)
        artifact.download(root=directory)
        try:
            artifact.verify(root=directory)
        except ValueError:
            if match_directory_exactly:
                artifact.checkout(root=directory)
            else:
                warnings.warn(
                    f"Artifact {full_artifact_name} content doesn't match exactly the content in directory {directory}."
                )
    return artifact


def construct_decorator(pass_function_args: bool = False):
    """
    A helper decorator to convert generator with contextlib.contextmanager like
    protocol to a decorator.

    :param pass_function_args: whether to pass function_args
        and function_kwargs keyword arguments to the generator.

    The example of generator with expected protocol:

    @construct_decorator()
    def decorator(function, log_time=True):
        logger = Logger()
        function_return_value = yield
        logger.log(function_return_value)
        if log_time:
            logger.log(time.time())

    @decorator(log_time=False)
    def hello(world=42):
        return world

    """

    def generator_decorator(generator):
        @functools.wraps(generator)
        def wrapped_generator(*generator_args, **generator_kwargs):
            def function_decorator(function):
                @functools.wraps(function)
                def wrapped_function(*function_args, **function_kwargs):
                    if pass_function_args:
                        generator_kwargs.update(
                            dict(
                                function_args=function_args,
                                function_kwargs=function_kwargs,
                            )
                        )
                    iterator = generator(function, *generator_args, **generator_kwargs)
                    next(iterator)
                    result = function(*function_args, **function_kwargs)
                    try:
                        iterator.send(result)
                    except StopIteration:
                        pass
                    return result

                return wrapped_function

            return function_decorator

        return wrapped_generator

    return generator_decorator


@construct_decorator()
def profile_to_file(function: FunctionType, out_file: io.TextIOBase = sys.stderr):
    profiler = line_profiler.LineProfiler()
    profiler.enable_by_count()
    profiler.add_function(function)
    yield
    profiler.print_stats(stream=out_file)


@construct_decorator()
def profile(function: FunctionType, logger: Logger = WandbLogger(), log_every=1):
    logger.register_function(function=function)
    yield
    logger.register_function_call(function=function, log_every=log_every)


def init_torch(debug: bool = True):
    if torch.cuda.is_available():
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        torch.cuda.init()
        torch.backends.cudnn.benchmark = True
    torch.autograd.set_detect_anomaly(mode=debug)


class CacheFlag(metaclass=Singleton):
    _enabled = [False]

    @property
    def enabled(self):
        return self._enabled[-1]

    def enter(self, enable: bool):
        assert isinstance(enable, bool), "Enable must be of bool type."
        self._enabled.append(enable)

    def exit(self):
        self._enabled.pop()


@contextlib.contextmanager
def cache(enable=True):
    """
    A contextmanager that together with switchable_cache decorator
    allows for more control over cache mechanics: functools.lru_cache
    will only be used on decorated functions inside enabled cache contextmanager.
    """
    CacheFlag().enter(enable=enable)
    yield
    CacheFlag().exit()


def switchable_cache(*args, **kwargs):
    """Decorator to mark functions that later can be used in cache contextmanager."""

    @functools.wraps(functools.lru_cache)
    def decorator(function):
        cached_function = functools.lru_cache(*args, **kwargs)(function)

        @functools.wraps(function)
        def wrapped_function(*function_args, **function_kwargs):
            if CacheFlag().enabled:
                return cached_function(*function_args, **function_kwargs)
            else:
                return function(*function_args, **function_kwargs)

        return wrapped_function

    return decorator


def filter_mro_by_classes_which_defined_function(
    cls: type, function_name: str
) -> "Generator[type, None, None]":
    for subclass in cls.mro():
        if not hasattr(subclass, function_name):
            break
        function = getattr(subclass, function_name)
        if type(function) is not types.FunctionType:
            raise ValueError(f"Attribute {function_name} is not a function.")
        if function.__qualname__.split(".")[0] == subclass.__name__:
            yield subclass


def clipped_tensor_string_representation(
    tensor: torch.Tensor, max_size: int = 50
) -> str:
    """Returns string representation of a tensor, clipped to max_size."""
    string = str(tensor)
    if len(string) >= max_size:
        string = string[: max_size // 2] + " ... " + string[-max_size // 2 :]
    string = re.sub(r"\s+", " ", string)
    return string



import warnings

import numpy as np
import scipy.sparse
import shap
import torch

from my_tools.models import (
    register_regularization_hook,
    StoppingMonitor,
    WandbLoggerMixin,
)
from my_tools.utils import torch_sparse_slice

from ..data import SparseDataset
from ..lit import LitRecommenderBase
from ..interface import RecommenderModuleBase, ExplanationMixin
from ..utils import plt_figure, wandb_plt_figure


class SLIM(RecommenderModuleBase, WandbLoggerMixin, ExplanationMixin):
    """
    The fitted model has sparse matrix W of shape [n_items, n_items],
    which it uses to predict ratings for users with feedback matrix
    A of shape [n_users, n_items] by calculating AW. It performs well,
    it's parameters are interpretable, and it performs inference quickly
    due to sparsity. It may be beneficial to add biases to this implementation.
    """

    def __init__(self, l2_coefficient=1.0e-4, l1_coefficient=1.0e-4, **kwargs):
        super().__init__(**kwargs, persistent_explicit=False)
        self.l2_coefficient = l2_coefficient
        self.l1_coefficient = l1_coefficient

        self.register_buffer(
            name="sparse_weight",
            tensor=torch.sparse_coo_tensor(size=(self.n_items, self.n_items)),
        )
        self.item_bias = torch.nn.Parameter(data=torch.zeros(self.n_items))
        self.dense_weight_slice = torch.nn.parameter.Parameter(torch.empty(0))

        self._sparse_values = torch.empty(0)
        self._sparse_indices = torch.empty(2, 0, dtype=torch.int32)

    def is_uninitialized(self):
        return self.dense_weight_slice.numel() == 0

    def init_dense_weight_slice(self, item_ids):
        dense_weight_slice = torch.empty(
            self.n_items, len(item_ids), device=self.dense_weight_slice.device
        )
        torch.nn.init.xavier_normal_(dense_weight_slice)
        dense_weight_slice = dense_weight_slice.abs()
        dense_weight_slice[item_ids, torch.arange(len(item_ids))] = 0
        self.dense_weight_slice = torch.nn.parameter.Parameter(dense_weight_slice)

    def transform_dense_slice_to_sparse(self, item_ids):
        sparse = self.dense_weight_slice.cpu().detach().to_sparse_coo()
        self._sparse_values = torch.cat([self._sparse_values, sparse.values()])
        rows, cols = sparse.indices()
        cols = item_ids[cols]
        indices = torch.stack([rows, cols])
        self._sparse_indices = torch.cat([self._sparse_indices, indices], dim=1)

        self.dense_weight_slice = torch.nn.parameter.Parameter(
            torch.empty(0, device=self.dense_weight_slice.device)
        )

        density = len(sparse.values()) / sparse.numel()
        self.log({"Items similarity matrix density": density})

    def finalize_training(self):
        self.sparse_weight = torch.sparse_coo_tensor(
            indices=self._sparse_indices,
            values=self._sparse_values,
            size=(self.n_items, self.n_items),
        )

    def clip_negative_parameter_values(self):
        self.dense_weight_slice = torch.nn.parameter.Parameter(
            self.dense_weight_slice.clip(0)
        )

    def training_forward(self, item_ids):
        assert len(item_ids) == self.dense_weight_slice.shape[1]
        dense_weight_slice = self.dense_weight_slice.clone()
        ratings = self.explicit.to(torch.float32) @ dense_weight_slice
        ratings = ratings.to_dense() + self.item_bias[item_ids]
        register_regularization_hook(
            dense_weight_slice, self.l2_coefficient, self.l1_coefficient
        )
        if dense_weight_slice.requires_grad:
            hook = SLIMHook(
                parameter=dense_weight_slice,
                fixed_row_id_in_each_col=item_ids,
            )
            dense_weight_slice.register_hook(hook)
        return ratings

    def forward(self, user_ids, item_ids):
        explicit = torch_sparse_slice(self.explicit, row_ids=user_ids).to(self.device)
        items_sparse_weight = torch_sparse_slice(
            self.sparse_weight, col_ids=item_ids
        ).to(self.device)
        ratings = explicit.to(torch.float32) @ items_sparse_weight
        ratings = ratings.to_dense() + self.item_bias[item_ids]
        return ratings

    def online_ratings(self, users_explicit):
        users_explicit = self.to_torch_coo(users_explicit)
        ratings = users_explicit.to(torch.float32).to(self.device) @ self.sparse_weight
        ratings = ratings.to_dense() + self.item_bias
        return ratings

    def explain_recommendations_for_user(
        self,
        user_id=None,
        user_explicit=None,
        n_recommendations=10,
        log=False,
        logging_prefix="",
        feature_names=None,
        n_background_samples_for_shap=1000,
    ):
        if feature_names is not None and len(feature_names) != self.n_items:
            raise ValueError(f"Feature names must be of length {self.n_items}")

        if user_id is not None:
            recommendations = self.recommend(
                user_ids=torch.IntTensor([user_id]), n_recommendations=n_recommendations
            )
            user_explicit = self.to_scipy_coo(self.explicit).tocsr()[user_id].toarray()
        else:
            recommendations = self.online_recommend(
                users_explicit=user_explicit, n_recommendations=n_recommendations
            )
            user_explicit = self.to_scipy_coo(user_explicit).toarray()

        background_samples_for_shap = self.to_scipy_coo(self.explicit).tocsr()[
            np.random.choice(
                np.arange(self.n_users),
                replace=False,
                size=min(n_background_samples_for_shap, self.n_users),
            )
        ]

        figures = []
        figure_context_manager = wandb_plt_figure if log else plt_figure
        for item_id in recommendations.squeeze(0).cpu().numpy():
            item_weight = torch_sparse_slice(self.sparse_weight, col_ids=[item_id])
            item_weight = item_weight.to_dense().numpy().squeeze(1)
            coefficient = self.item_bias[item_id].item()
            explainer = shap.explainers.Linear(
                model=(item_weight, coefficient),
                masker=background_samples_for_shap,
                feature_names=feature_names,
            )
            shap_values = explainer(user_explicit)
            title = (
                logging_prefix
                + f" Item {item_id if feature_names is None else feature_names[item_id]}"
            )
            with figure_context_manager(title=title) as figure:
                shap.waterfall_plot(shap_values[0])
            figures.append(figure)

        return figures


class SLIMHook:
    """
    Gradient hook intended to clip gradient to enforce the positive
    parameter values and zero diagonal regularization restrictions.
    It would be easier to clip values after optimizer step, but then
    it would mess up optimizer's running gradient mean estimation
    and parameter weighting. We still need to clip parameters after
    the step though as the gradient step may still make positive
    parameters negative.
    """

    def __init__(self, parameter, fixed_row_id_in_each_col):
        self.fixed_row_id_in_each_col = fixed_row_id_in_each_col
        self.parameter = parameter.clone().detach()

    def soft_positive_regularization(self, grad):
        grad[(self.parameter == 0) & (0 < grad)] = 0

    def zero_parameter_diagonal_preservation(self, grad):
        grad[
            self.fixed_row_id_in_each_col,
            torch.arange(len(self.fixed_row_id_in_each_col)),
        ] = 0

    def __call__(self, grad):
        grad = grad.clone().detach()
        self.soft_positive_regularization(grad)
        self.zero_parameter_diagonal_preservation(grad)
        return grad


class SLIMDataset(SparseDataset):
    def __init__(
        self,
        explicit: scipy.sparse.csr_matrix,
        explicit_val: scipy.sparse.csr_matrix,
    ):
        super().__init__(explicit)
        self.explicit_val = explicit_val

    def __len__(self):
        return self.explicit.shape[1]

    def __getitem__(self, indices):
        item = super().__getitem__(indices)
        explicit_val = torch_sparse_slice(
            self.explicit_val, item["user_id"], item["item_id"]
        )
        item.update(explicit_val=self.pack_sparse_tensor(explicit_val))
        return item


class SLIMRecommender(LitRecommenderBase):
    def __init__(
        self,
        patience=0,
        min_delta=0,
        check_val_every_n_epoch=1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.save_hyperparameters(ignore=list(kwargs.keys()))
        self.stopping_monitor = None
        self.current_batch = None
        self.batch_is_fitted = False
        self.item_dataloader_iter = None

    @property
    def class_candidates(self):
        return super().class_candidates + [SLIM]

    def setup(self, stage=None):
        super().setup(stage)
        if stage == "fit":
            if self.trainer.limit_val_batches != 0:
                warnings.warn(
                    "In this implementation when training model is iterating over items, "
                    "and the user-wise metrics cannot be calculated while training. "
                    "So validation should be manually disabled, and the model will validate "
                    "when it is fitted."
                )
            self.stopping_monitor = StoppingMonitor(
                self.hparams["patience"], self.hparams["min_delta"]
            )
            self.current_batch = None
            self.batch_is_fitted = False
            slim_dataset = SLIMDataset(self.train_explicit(), self.val_explicit())
            item_dataloader = self.build_dataloader(
                dataset=slim_dataset, sampler_type="item"
            )
            self.item_dataloader_iter = iter(item_dataloader)

    def train_dataloader(self):
        try:
            self.current_batch = next(self.item_dataloader_iter)
        except StopIteration:
            self.trainer.should_stop = True
            yield self.current_batch

        while not self.batch_is_fitted:
            yield self.current_batch

    def on_train_batch_start(self, batch, batch_idx):
        if self.batch_is_fitted:
            self.batch_is_fitted = False
            return -1
        if self.model.is_uninitialized():
            self.model.init_dense_weight_slice(item_ids=batch["item_id"])

    def training_step(self, batch, batch_idx):
        ratings = self.model.training_forward(item_ids=batch["item_id"])
        train_loss = self.loss(explicit=batch["explicit"], model_ratings=ratings)
        self.log("train_loss", train_loss)

        if batch_idx % self.hparams["check_val_every_n_epoch"] == 0:
            val_loss = self.loss(explicit=batch["explicit_val"], model_ratings=ratings)
            self.log("val_loss", val_loss)

            if self.stopping_monitor.is_time_to_stop(val_loss):
                self.batch_is_fitted = True
                self.model.transform_dense_slice_to_sparse(
                    item_ids=self.current_batch["item_id"]
                )

        return train_loss

    def on_train_batch_end(self, outputs, batch, batch_idx):
        self.model.clip_negative_parameter_values()
        # Update parameters passed to optimizer.
        self.trainer.optimizers = [self.configure_optimizers()]

    def on_train_end(self):
        self.model.finalize_training()
        self.trainer.validate(self)
        super().on_train_end()



from typing import TypeVar, Callable, TYPE_CHECKING

import torch

if TYPE_CHECKING:
    from .utils import SparseTensor


Distance = TypeVar(
    "Distance", bound=Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
)


def cosine_distance(
    left: torch.Tensor or "SparseTensor", right: torch.Tensor or "SparseTensor"
) -> torch.Tensor:
    """
    Given two matrices: left of shape [n, d] and right of shape [m, d],
    containing in rows vectors from d-dimensional space, returns matrix S
    of shape [n, m], where S[i, j] = cosine distance between left[i] and right[j].
    """

    def norm(tensor):
        if tensor.is_sparse:
            return torch.sqrt(torch.sparse.sum(tensor**2, dim=1)).to_dense()
        else:
            return torch.sqrt(torch.sum(tensor**2, dim=1))

    left = left.to(torch.float32)
    right = right.to(torch.float32)
    similarity = (left @ right.transpose(0, 1)).to_dense()
    similarity /= norm(left).reshape(-1, 1)
    similarity /= norm(right).reshape(1, -1)
    return 1 - similarity


def weighted_average(tensor, weights):
    """
    Given tensor T of shape [n, m] and weights W of shape [n],
    returns weighted average A of shape [m]:
    A[i] = \sum_j T[j, i] * W[j]
    """
    if (weights < 0).any():
        raise ValueError("Weights must be positive.")
    if (weights == 0).all():
        weights = torch.ones_like(weights)
    weights /= weights.sum()
    return tensor.T @ weights


def safe_log(tensor: torch.Tensor) -> torch.Tensor:
    """Returns 0 if x == 0, else log(x)"""
    tensor[tensor == 0] = 1
    return torch.log(tensor)


def kl_divergence(p: torch.Tensor, q: torch.Tensor) -> torch.Tensor:
    if (p < 0).any() or not torch.isclose(p.sum(), torch.tensor(1.0)):
        raise ValueError("p must be a probability distribution")
    if (q < 0).any() or not torch.isclose(q.sum(), torch.tensor(1.0)):
        raise ValueError("q must be a probability distribution")
    return (p * (safe_log(p) - safe_log(q))).sum()


def pairwise_difference(left, right):
    """
    Given two matrices left, right of shape (n, m),
    returns matrix D of shape (n, m, m) such that
    D[u, i, j] = left[u, i] - right[u, j]
    """
    return left[:, :, None] - right[:, None, :]



import abc
import warnings
from abc import ABC
from typing import TYPE_CHECKING, Any

import torch
import wandb

from .maths import Distance, cosine_distance, weighted_average
from .utils import profile, Timer
from my_tools.utils import to_torch_coo, to_scipy_coo, torch_sparse_slice

if TYPE_CHECKING:
    from scipy.sparse import spmatrix
    from my_tools.utils import SparseTensor, Pickleable


class RecommenderModuleInterface(torch.nn.Module, abc.ABC):
    @property
    @abc.abstractmethod
    def device(self):
        """Returns the device on which the module resides at the moment."""

    @abc.abstractmethod
    def forward(
        self, user_ids: torch.IntTensor, item_ids: torch.IntTensor
    ) -> torch.FloatTensor:
        ...

    @abc.abstractmethod
    def recommend(
        self, user_ids: torch.IntTensor, n_recommendations: int = None
    ) -> torch.IntTensor:
        """Returns recommended item ids."""

    @abc.abstractmethod
    def online_ratings(self, explicit: "SparseTensor" or "spmatrix") -> torch.Tensor:
        """Generate ratings for new users defined by their explicit feedback"""

    @abc.abstractmethod
    def online_recommend(
        self,
        explicit: "SparseTensor" or "spmatrix",
        n_recommendations: int = None,
    ) -> torch.IntTensor:
        """Generate recommendations for new users defined by their explicit feedback."""


class RecommenderModuleBase(RecommenderModuleInterface, ABC):
    explicit: torch.Tensor

    def __init__(
        self,
        n_users: int,
        n_items: int,
        explicit: "SparseTensor" or "spmatrix" or None = None,
        persistent_explicit=False,
        invalid_recommendation_mark: int = -1,
    ):
        super().__init__()
        self.n_users = n_users
        self.n_items = n_items
        if explicit is not None:
            explicit = to_torch_coo(explicit)
        self.register_buffer(
            name="explicit", tensor=explicit, persistent=persistent_explicit
        )
        self.invalid_recommendation_mark = invalid_recommendation_mark
        self.to_torch_coo = to_torch_coo
        self.to_scipy_coo = to_scipy_coo

    @property
    def device(self):
        if self.explicit is not None:
            return self.explicit.device
        for parameter in self.parameters():
            return parameter.device

    @Timer()
    @profile()
    def ratings_to_filtered_recommendations(
        self, explicit, ratings, n_recommendations=None
    ):
        n_recommendations = n_recommendations or self.n_items
        ratings = torch.where(explicit.to_dense() > 0, -torch.inf, ratings)
        item_ratings, item_ids = torch.topk(input=ratings, k=n_recommendations)
        item_ids[item_ratings == -torch.inf] = self.invalid_recommendation_mark
        self.check_invalid_recommendations(recommendations=item_ids, warn=False)
        return item_ids

    def check_invalid_recommendations(self, recommendations, warn: bool):
        invalid_items_mask = (recommendations == self.invalid_recommendation_mark).to(
            torch.float32
        )
        users_without_enough_recommendations = invalid_items_mask.any(1).to(
            torch.float32
        )
        if wandb.run is not None:
            wandb.log(
                {
                    "sanity_check/average_n_invalid_recommendations_per_user": invalid_items_mask.mean(),
                    "sanity_check/not_enough_recommendations_frequency": users_without_enough_recommendations.mean(),
                }
            )
        if warn and any(users_without_enough_recommendations):
            warning_message = (
                f"Model failed to generate requested number of recommendations for users.\n"
                f"Model was asked for {recommendations.shape[1]} recommendations among "
                f"{self.n_items} items, but {users_without_enough_recommendations.sum()} "
                f"users out of batch total {recommendations.shape[0]} did not get enough recommendations."
            )
            warnings.warn(warning_message, category=RuntimeWarning)

    def recommend(self, user_ids, n_recommendations=None):
        with torch.inference_mode():
            ratings = self(user_ids=user_ids, item_ids=torch.arange(self.n_items))
        users_explicit = torch_sparse_slice(self.explicit, row_ids=user_ids).to(
            self.device
        )
        recommendations = self.ratings_to_filtered_recommendations(
            explicit=users_explicit,
            ratings=ratings,
            n_recommendations=n_recommendations,
        )
        return recommendations

    def online_nn_ratings(
        self,
        explicit: "SparseTensor" or spmatrix,
        distance: Distance or None = cosine_distance,
        n_neighbours: int = 10,
    ):
        explicit = self.to_torch_coo(explicit).to(self.device).detach().clone()
        distances = distance(explicit, self.explicit.detach().clone())
        neighbors_distances, nearest_user_ids = torch.topk(
            distances, k=n_neighbours, largest=False
        )
        ratings = torch.empty(
            *explicit.shape,
            dtype=torch.float32,
            device=self.device,
        )
        for i, (user_ids, distances_to_users) in enumerate(
            zip(nearest_user_ids, neighbors_distances)
        ):
            with torch.inference_mode():
                neighbours_ratings = self(
                    user_ids=user_ids, item_ids=torch.arange(self.n_items)
                )
            user_ratings = weighted_average(
                tensor=neighbours_ratings,
                weights=(distances_to_users + torch.finfo().eps) ** -1,
            )
            ratings[i] = user_ratings
        return ratings

    def online_ratings(self, users_explicit):
        """
        The fallback method is based on nearest neighbours,
        as it can be applied to any model, but may give suboptimal
        recommendations. Feel free to overwrite this
        method if model natively supports online ratings.
        """
        return self.online_nn_ratings(users_explicit)

    def forward(self, user_ids, item_ids):
        users_explicit = torch_sparse_slice(self.explicit, row_ids=user_ids)
        ratings = self.online_ratings(users_explicit=users_explicit)
        return ratings[:, item_ids]

    def online_recommend(self, users_explicit, n_recommendations=None):
        users_explicit = self.to_torch_coo(users_explicit).to(self.device)
        ratings = self.online_ratings(users_explicit)
        recommendations = self.ratings_to_filtered_recommendations(
            explicit=users_explicit,
            ratings=ratings,
            n_recommendations=n_recommendations,
        )
        return recommendations

    def get_extra_state(self) -> "Pickleable":
        """
        Whatever extra non-tensor parameters need to be saved for inference,
        will be taken here on checkpoint save and passed to load method on checkpoint load.
        This is torch.nn.Module method, this stub is just for clarity.
        """

    def set_extra_state(self, state: "Pickleable") -> None:
        """
        Init module from whatever was returned by getter to be ready for inference.
        This is torch.nn.Module method, this stub is just for clarity.
        """


class FitExplicitInterfaceMixin:
    @abc.abstractmethod
    def fit(self) -> None:
        """Fit to self.explicit feedback matrix."""


class ExplanationMixin:
    @abc.abstractmethod
    def explain_recommendations_for_user(
        self,
        user_id: int = None,
        user_explicit: "SparseTensor" = None,
        n_recommendations=10,
        log: bool = False,
        logging_prefix: str = "explanation/",
    ) -> Any:
        """
        Return any data that will aid in understanding why
        the model recommends the way it does.
        :param user_id: the id of user, for whom the
        recommendations are generated, if he has one
        :param user_explicit: the user's feedback of
        shape [1, self.n_items], if he doesn't have an id
        :param n_recommendations: number of recommendations
        :param log: whether to log explanations to wandb
        :param logging_prefix: the prefix for log entries
        """


class UnpopularRecommenderMixin:
    """
    I want to recommend items which the users haven't previously seen,
    so given probability p_ui that user has seen item, I want to decrease
    the expected rating r_ui by an amount depending on that probability:
    predicted_relevance_ui = r_ui - g(p_ui),
    where g is a non-decreasing function.

    For convenience and without much complexity sacrifice, p_ui can be
    factorized as p_u * p_i, and p_u and p_i can be
    estimated as sample frequencies.

    The implementation below uses g(x) = unpopularity_coef * log(p_ui).
    """

    unpopularity_coef: torch.Tensor
    users_activity: torch.Tensor
    items_popularity: torch.Tensor

    def init_unpopular_recommender_mixin(
        self: RecommenderModuleBase, unpopularity_coef=1e-3
    ):
        self.register_buffer(
            name="unpopularity_coef", tensor=torch.tensor(unpopularity_coef)
        )
        self.register_buffer(name="users_activity", tensor=torch.zeros(self.n_users))
        self.register_buffer(name="items_popularity", tensor=torch.zeros(self.n_items))

    def fit_unpopular_recommender_mixin(self: RecommenderModuleBase):
        self.users_activity = torch.tensor(
            (self.to_scipy_coo(self.explicit) > 0).mean(1).A.squeeze(1)
        )
        self.items_popularity = torch.tensor(
            (self.to_scipy_coo(self.explicit) > 0).mean(0).A.squeeze(0)
        )

    def probability_that_user_has_seen_item(self, users_activity):
        return torch.einsum("u, i -> ui", users_activity, self.items_popularity)

    def additive_rating_offset(self, users_activity):
        p_ui = self.probability_that_user_has_seen_item(users_activity=users_activity)
        eps = torch.finfo().eps
        return -self.unpopularity_coef * torch.log(torch.FloatTensor(eps + p_ui))

    @property
    def mean_user_activity(self):
        """Convenience method."""
        return self.users_activity.mean(dim=0, keepdims=True)


class RecommendingLossInterface:
    def __init__(self, explicit):
        pass

    @abc.abstractmethod
    def __call__(
        self,
        explicit: "SparseTensor",
        model_ratings: torch.FloatTensor = None,
        model: "RecommenderModuleBase" = None,
        user_ids: torch.IntTensor = None,
        item_ids: torch.IntTensor = None,
    ) -> torch.FloatTensor:
        ...



from .data import MovieLensMixin
from ..models import als, baseline, cat, mf, slim
from ..lit import NonGradientRecommenderMixin, LitRecommenderBase
from . import cat as movielens_cat


class MovieLensNonGradientRecommenderMixin(NonGradientRecommenderMixin):
    @property
    def module_candidates(self):
        return super().module_candidates + [
            als,
            baseline,
            cat,
            movielens_cat,
        ]


class MovieLensRecommender(MovieLensMixin, LitRecommenderBase):
    pass


class MovieLensNonGradientRecommender(
    MovieLensNonGradientRecommenderMixin, MovieLensRecommender
):
    pass


class MovieLensMFRecommender(mf.MFRecommender, MovieLensRecommender):
    pass


class MovieLensSLIMRecommender(slim.SLIMRecommender, MovieLensRecommender):
    pass
